__all__ = [
    "df_channel",
    "df_diamonds",
    "df_ruff",
    "df_stang",
    "df_stang_wide",
    "df_trajectory_full",
    "df_trajectory_windowed",
    "df_shewhart",
]

import os
from pandas import read_csv
from pathlib import Path


path_this = Path(__file__)
path_grama = path_this.parents[1]

# Stang (wdie form)
df_stang_wide = read_csv(Path(path_grama / "data/stang.csv"))

# Stang (tidy form)
df_stang = read_csv(Path(path_grama / "data/stang_long.csv"))

# Stang (tidy form)
df_diamonds = read_csv(Path(path_grama / "data/diamonds.csv"))

# Ruff (tidy form)
df_ruff = read_csv(Path(path_grama / "data/ruff.csv"))

# Trajectories
df_trajectory_full = read_csv(Path(path_grama / "data/trajectory_full.csv"))
df_trajectory_windowed = read_csv(Path(path_grama / "data/trajectory_windowed.csv"))

# Shewhart
df_shewhart = read_csv(Path(path_grama / "data/shewhart1931-table3.csv"))

# PSAAP II IJMF (2020) data
df_channel = read_csv(Path(path_grama / "data/psaap.csv"))
df_channel.rename(
    columns=dict(
        # xi="x",
        x="xi",
        # H="W",
        W="H",
        # U="U_0",
        U_0="U",
        # T_0="T_f",
        T_f="T_0",
        # cp_p="C_pv",
        C_pv="cp_p",
        # cp_f="C_fp",
        C_fp="cp_f",
        # nu_f="mu_f",
        mu_f="nu_f",
        # Q_abs="eps_p",
        eps_p="Q_abs",
        # h_p="h",
        h="h_p",
    ),
    inplace=True,
)
# Compute thermal diffusivity from initial temperature
df_channel["alpha_f"] = (
      1.862e1
    + 1.327e-1 * df_channel.T_0
    + 1.026e-4 * df_channel.T_0**2
    - 5.270e-9 * df_channel.T_0**3
) * 1e-6
# Back-calculate the proper number density
df_channel["n"] = df_channel.N_p / df_channel.H**2 / df_channel.L



all = [
    "eval_pnd",
    "ev_pnd",
    "pareto_min_rel",
    "make_proposal_sigma",
    "rprop",
    "dprop",
    "approx_pnd"
]

from grama import add_pipe, Model, tf_md
from grama.eval_defaults import invariants_eval_df, invariants_eval_model

from numpy import array, diag, dot, ones, mean, zeros
from numpy import any as npany
from numpy import all as npall
from numpy import min as npmin
from numpy import max as npmax
from numpy import sum as npsum
from numpy.linalg import norm
from numpy.random import choice, multivariate_normal
from numpy.random import seed as set_seed
from pandas import DataFrame
from scipy.linalg import svd
from scipy.stats import multivariate_normal as mvnorm
from toolz import curry


@curry
def eval_pnd(model, df_train, df_test, signs, n=int(1e4), seed=None, append=True, \
    mean_prefix="_mean", sd_prefix="_sd"):
    """Approximate the probability non-dominated (PND)

    Approximates the probability non-dominated (PND) for a set of training points given a fitted probabilistic model. Used to rank a set of candidates in the context of multiobjective optimization.

    Args:
        model (gr.model): predictive model to evaluate
        df_train (DataFrame): dataframe with training data
        df_test (DataFrame): dataframe with test data
        signs (dict): dict with the variables you would like to use and
            minimization or maximization parameter for each
        append (bool): Append df_test to pnd algorithm outputs

    Kwargs:
        n (int): Number of draws for importance sampler
        seed (int): declarble seed value for reproducibility

    Returns:
        DataFrame: Results of predictive model going through a PND algorithm.
        Conatians both values and their scores.

    References:
        del Rosario, Zachary, et al. "Assessing the frontier: Active learning, model accuracy, and multi-objective candidate discovery and optimization." The Journal of Chemical Physics 153.2 (2020): 024112.

    Examples::

        import grama as gr

        ## Define a ground-truth model
        md_true = gr.make_pareto_random()
        df_data = (
            md_true
            >> gr.ev_sample(n=2e3, seed=101, df_det="nom")
        )
        ## Generate test/train data
        df_train = (
            df_data
            >> gr.tf_sample(n=10)
        )

        df_test = (
            df_data
            >> gr.anti_join(
                df_train,
                by = ["x1","x2"]
            )
            >> gr.tf_sample(n=200)
        )
        ## Fit a model to training data
        md_fit = (
            df_train
            >> gr.ft_gp(
                var=["x1","x2"]
                out=["y1","y2"]
            )
        )
        ## Rank training points by PND algorithm
        df_pnd = (
            md_fit
            >> gr.ev_pnd(
                df_train,
                df_test,
                signs = {"y1":1, "y2":1},
                seed = 101
            )
            >> gr.tf_arrange(gr.desc(DF.pr_scores))
        )

    """
    invariants_eval_model(model)
    invariants_eval_df(df_train, arg_name="df_train")
    invariants_eval_df(df_test, arg_name="df_test")


    # Check content
    if len(model.out)/2 < 2:
        raise ValueError('Given Model needs multiple outputs')

    if not set(model.var).issubset(set(df_train.columns)):
        raise ValueError("model.var must be subset of df_train.columns")

    if not set(model.var).issubset(set(df_test.columns)):
        raise ValueError("model.var must be subset of df_test.columns")

    for key in signs.keys():
        if key+mean_prefix not in model.out:
            raise ValueError(f"signs.{key} implies output {key+mean_prefix}, which is not found in provided md.out")
        if key+sd_prefix not in model.out:
            raise ValueError(f"signs{key} implies output {key+sd_prefix}, which is not found in provided sd.out")

    ## Compute predictions and predicted uncertainties
    df_pred = (
        df_test
        >> tf_md(md=model)
    )

    ## Setup for reshaping
    means = []
    sds = []
    columns = df_train.columns.values
    length = int(len(signs.keys()))
    outputs = [key for key in signs.keys() if key in columns]
    signs = [value for value in signs.values()]

    ## append mean and sd prefixes
    for i, value in enumerate(outputs):
        means.append(value+mean_prefix)
        sds.append(value+sd_prefix)

    ## Remove extra columns from df_test
    df_pred = df_pred[means + sds]

    ## Reshape data for PND algorithm
    X_pred = df_pred[means].values      # Predicted response values
    X_sig = df_pred[sds].values         # Predictive uncertainties
    X_train = df_train[outputs].values  # Training

    ### Create covariance matrices
    X_cov = zeros((X_sig.shape[0], length, length))
    for l in range(length):
        for i in range(X_sig.shape[0]):
            X_cov[i, l, l] = X_sig[i, l]

    ### Apply pnd
    pr_scores, var_values = approx_pnd(
        X_pred,
        X_cov,
        X_train,
        signs = signs,
        n = n,
        seed = seed
    )

    ### Package outputs
    df_pnd = DataFrame(
        {
            "pr_scores": pr_scores,
            "var_values": var_values,
        }
    )

    if append:
        return df_test.reset_index(drop=True).merge(df_pnd, left_index=True, right_index=True)
    return df_pnd

ev_pnd = add_pipe(eval_pnd)


# Relative Pareto frontier calculation
def pareto_min_rel(X_test, X_base=None):
    r"""Determine if rows in X_test are optimal, compared to X_base

    Finds the Pareto-efficient test-points that minimize the column values,
    relative to a given set of base-points.

    Args:
        X_test (2d numpy array): Test point observations; rows are observations, columns are features
        X_base (2d numpy array): Base point observations; rows are observations, columns are features

    Returns:
        array of boolean values: Indicates if test observation is Pareto-efficient, relative to base points

    References:
        Owen *Monte Carlo theory, methods and examples* (2013)
    """
    # Compute Pareto points
    is_efficient = ones(X_test.shape[0], dtype=bool)

    if X_base is None:
        for i, x in enumerate(X_test):
            is_efficient[i] = npall(npany(X_test[:i] > x, axis=1)) and npall(
                npany(X_test[i + 1 :] > x, axis=1)
            )
    else:
        for i, x in enumerate(X_test):
            is_efficient[i] = not (
                npany(npall(x >= X_base, axis=1))
                and npany(npany(x >  X_base, axis=1))
            )

    return is_efficient

# Floor an array of variances
def floor_sig(sig, sig_min=1e-8):
    r"""Floor an array of variances
    """
    sig_floor = npmin([norm(sig), sig_min])
    return list(map(lambda s: npmax([s, sig_floor]), sig))

# Estimate parameters for PND importance distribution
def make_proposal_sigma(X, idx_pareto, X_cov):
    r"""Estimate parameters for PND importance distribution

    Args:
        X (2d numpy array): Full set of response values
        idx_pareto (boolean array): Whether each response value is a Pareto point
        X_cov (iterable of 2d numpy array): Predictive covariance matrices

    Preconditions:
        X.shape[0] == len(idx_pareto)
        X_cov[i].shape[0] == X.shape[1] for all valid i
        X_cov[i].shape[1] == X.shape[1] for all valid i

    Returns:
        2d numpy array: Common covariance matrix for PND importance distribution
    """
    n_pareto = len(idx_pareto)
    n_dim = X.shape[0]
    X_mean = mean(X[idx_pareto, :], axis=0)

    ## If sufficient, use Pareto points only
    if n_pareto > n_dim:
        U, sig, Vh = svd(X[idx_pareto, :] - X_mean)
    ## If insufficient, use full distribution
    else:
        U, sig, Vh = svd(X - X_mean)

    ## Find largest predictive covariance component to avoid dangerously light tails
    sig_min = npmax(X_cov)
    ## Apply heuristic based on Owen "Monte Carlo", Exercise 9.7
    sig_min = ((5/4)**1/2) * sig_min

    ## Floor the variances
    sig = floor_sig(sig, sig_min=sig_min)

    ## TODO: Develop better heuristic
    Sigma = dot(Vh.T, dot(diag(sig), Vh)) / n_pareto

    return Sigma

# Draw from mixture distribution
def rprop(n, Sigma, X_means, seed=None):
    r"""Draw a sample from gaussian mixture distribution

    Draw a sample from a gaussian mixture distribution with a common covariance
    matrix and different means.

    Args:
        n (int): Number of observations in sample
        Sigma (2d numpy array): Common covariance matrix for mixture distribution
        X_means (2d numpy array): Means for mixture distribution

    Preconditions:
        Sigma.shape[0] == X_means.shape[1]
        Sigma.shape[1] == X_means.shape[1]

    Returns:
        2d numpy array: Sample of observations
    """
    if seed is not None:
        set_seed(seed)

    n_pareto, n_dim = X_means.shape
    X_sample = zeros((n, n_dim))

    for i in range(n):
        idx = choice(n_pareto)
        X_sample[i, :] = multivariate_normal(
            mean=X_means[idx, :],
            cov=Sigma,
            size=1,
        )

    return X_sample

# Calculate density based on proposal distribution
def dprop(X, Sigma, X_means):
    r"""Evaluate the PDF of a mixture proposal distribution

    Evaluate the PDF of a gaussian mixture distribution with a common covariance
    matrix and different means.

    Args:
        X (2d numpy array): Observations for which to evaluate the density
        Sigma (2d numpy array): Common covariance matrix for mixture distribution
        X_means (2d numpy array): Means for mixture distribution

    Preconditions:
        X.shape[1] == X_means.shape[1]
        Sigma.shape[0] == X_means.shape[1]
        Sigma.shape[1] == X_means.shape[1]

    Returns:
        numpy array: Density values
    """
    n, n_dim = X.shape
    n_comp = X_means.shape[0]
    w = 1 / n_comp           # Equal weighting of mixture components

    L = zeros((n_comp, n))
    dist = mvnorm(cov=Sigma)
    for i in range(n_comp):
        L[i, :] = dist.pdf(X - X_means[i])

    return npsum(L, axis=0) * w

# Approximate PND via IS
def approx_pnd(X_pred, X_cov, X_train, signs, n=int(1e4), seed=None):
    r"""Approximate the PND via mixture importance sampling

    Approximate the probability non-dominated (PND) for a set of predictive
    points using a mixture importance sampling approach. Predictive points are
    assumed to have predictive gaussian distributions (with specified mean and
    covariance matrix).

    Args:
        X_pred (2d numpy array): Predictive values
        X_cov (iterable of 2d numpy arrays): Predictive covariance matrices
        X_train (2d numpy array): Training values, used to determine existing Pareto frontier
        signs (numpy array of +/-1 values): Array of optimization signs: {-1: Minimize, +1 Maximize}

    Kwargs:
        n (int): Number of draws for importance sampler
        seed (int): Seed for random state

    Returns:
        pr_scores (array): Estimated PND values
        var_values (array): Estimated variance values

    References:
        Owen *Monte Carlo theory, methods and examples* (2013)

    """
    ## Setup
    X_wk_train = -X_train * signs
    X_wk_pred = -X_pred * signs
    n_train, n_dim = X_train.shape
    n_pred = X_pred.shape[0]

    ## Find the training Pareto frontier
    idx_pareto = pareto_min_rel(X_wk_train)
    n_pareto = len(idx_pareto)
    ## Sample the mixture points
    Sig_mix = make_proposal_sigma(X_wk_train, idx_pareto, X_cov)
    X_mix = rprop(n, Sig_mix, X_wk_train[idx_pareto, :], seed=seed)
    ## Take non-dominated points only
    idx_ndom = pareto_min_rel(X_mix, X_base=X_wk_train[idx_pareto, :])
    X_mix = X_mix[idx_ndom, :]

    ## Evaluate the Pr[non-dominated]
    d_mix = dprop(X_mix, Sig_mix, X_wk_train[idx_pareto, :])
    pr_scores = zeros(n_pred)
    var_values = zeros(n_pred)
    for i in range(n_pred):
        dist_test = mvnorm(mean=X_wk_pred[i], cov=X_cov[i])
        w_test = dist_test.pdf(X_mix) / d_mix
        # Owen (2013), Equation (9.3)
        pr_scores[i] = npsum(w_test) / n
        # Owen (2013), Equation (9.5)
        var_values[i] = npsum( (w_test - pr_scores[i])**2 ) / n

    return pr_scores, var_values



from __future__ import print_function
import numpy as np
import scipy.linalg

__all__ = ['gauss']

def gauss(N, a = 0, b = 1):
	r""" Gauss-Legendre quadrature rule



	Parameters
	----------
	N: int
		number of samples to use in quadrature rule
	a: float, default 0
		left endpoint
	b: float, default 1
		right endpoint

	Returns
	-------
	x: np.ndarray
		locations at which to sample
	w: np.ndarray
		weights

	Notes
	-----
	This code is adapted from Mark Embree's `gaussab.m`,
	a modification of `gauss.m` from Trefethen's Spectral Methods in Matlab
	"""
	a, b = float(a), float(b)

	beta = 0.5/np.sqrt(1 - ((np.arange(2,2*N, 2, dtype = float)**(-2) )))
	T = np.diag(beta, 1) + np.diag(beta, -1)

	# Because eigenvalues come sorted in increasing order, there is no need to sort
	ew, ev = scipy.linalg.eigh(T)
	x = ew
	w = 2*ev[0,:]**2

	# Mark's modifications
	x = a + (b - a)/2.*(1.+x)
	w = ((b-a)/2.)*w

	return x, w



""" Sequential Linear Program
"""
from __future__ import print_function
import numpy as np
import scipy.optimize
import warnings

from .domains import UnboundedDomain
from .gn import trajectory_linear

class InfeasibleException(Exception):
	pass

class UnboundedException(Exception):
	pass


def sequential_lp(f, x0, jac, search_constraints = None,
	norm = 2, trajectory = trajectory_linear, obj_lb = None, obj_ub = None,
	constraints = None, constraint_grads = None, constraints_lb = None, constraints_ub = None,
	maxiter = 100, bt_maxiter = 50, domain = None,
	tol_dx = 1e-10, tol_obj = 1e-10,  verbose = False, **kwargs):
	r""" Solves a nonlinear optimization problem by a sequence of linear programs


	Given the optimization problem

	.. math::

		\min{\mathbf{x} \in \mathbb{R}^m} &\ \| \mathbf{f}(\mathbf{x}) \|_p \\
		\text{such that} & \  \text{lb} \le \mathbf{f} \le \text{ub} \\
			& \ \text{constraint_lb} \le \mathbf{g} \le \text{constraint_ub}

	this function solves this problem by linearizing both the objective and constraints
	and solving a sequence of disciplined convex problems.


	Parameters
	----------
	norm: [1,2, np.inf, None, 'hinge']
		If hinge, sum of values of the objective exceeding 0.


	References
	----------
	.. FS89


	"""
	assert norm in [1,2,np.inf, None, 'hinge'], "Invalid norm specified."

	if search_constraints is None:
		search_constraints = lambda x, p: []

	if domain is None:
		domain = UnboundedDomain(len(x0))

	if constraints is None:
		constraints = []
	if constraint_grads is None:
		constraint_grads = []

	assert len(constraints) == len(constraint_grads), "Must provide same number of constraints as constraint gradients"

	if constraints_lb is None:
		constraints_lb = -np.inf*np.ones(len(constraints))

	if constraints_ub is None:
		constraints_ub = np.inf*np.ones(len(constraints))


	# The default solver for 1/inf-norm doesn't converge sharp enough, but ECOS does.
	if 'solver' not in kwargs:
		kwargs['solver'] = 'ECOS'


	if norm in [1,2, np.inf]:
		objfun = lambda fx: np.linalg.norm(fx, ord = norm)
	elif norm == 'hinge':
		objfun = lambda fx: np.sum(np.maximum(fx, 0))
	else:
		objfun = lambda fx: float(fx)

	# evalutate KKT norm
	def kkt_norm(fx, jacx):
		kkt_norm = np.nan
		if norm == np.inf:
			# TODO: allow other constraints into the solution
			t = objfun(fx)
			obj_grad = np.zeros(len(x)+1)
			obj_grad[-1] = 1.
			con = np.hstack([fx - t, -fx -t])
			con_grad = np.zeros((2*len(fx),len(x)+1))
			con_grad[:len(fx),:-1] = jacx
			con_grad[:len(fx),-1] = -1.
			con_grad[len(fx):,:-1] = -jacx
			con_grad[len(fx):,-1] = -1.

			# Find the active constraints (which have non-zero Lagrange multipliers)
			I = np.abs(con) < 1e-10
			lam, kkt_norm = scipy.optimize.nnls(con_grad[I,:].T, -obj_grad)
		elif norm == 1:
			t = np.abs(fx)
			obj_grad = np.zeros(len(x) + len(fx))
			obj_grad[len(x):] = 1.
			con = np.hstack([fx - t, -fx-t])
			con_grad = np.zeros((2*len(fx), len(x)+len(fx)))
			con_grad[:len(fx),:len(x)] = jacx
			con_grad[:len(fx),len(x):] = -1.
			con_grad[len(fx):,:len(x)] = -jacx
			con_grad[len(fx):,len(x):] = -1.
			I = np.abs(con) == 0.
			lam, kkt_norm = scipy.optimize.nnls(con_grad[I,:].T, -obj_grad)

		elif norm == 2:
			kkt_norm = np.linalg.norm(jacx.T.dot(fx))

		# TODO: Should really orthogonalize against unallowed search directions
		#err = con_grad[I,:].T.dot(lam) + obj_grad
		#print err

		return kkt_norm

	# Start optimizaiton loop
	x = np.copy(x0)
	try:
		fx = np.array(f(x))
	except TypeError:
		fx = np.array([fi(x) for fi in f]).reshape(-1,)

	objval = objfun(fx)

	try:
		jacx = jac(x)
	except TypeError:
		jacx = np.array([jaci(x) for jaci in jac]).reshape(len(fx), len(x))

	if verbose:
		print('iter |     objective     |  norm px | TR radius | KKT norm | violation |')
		print('-----|-------------------|----------|-----------|----------|-----------|')
		print('%4d | %+14.10e |          |           | %8.2e |           |' % (0, objval, kkt_norm(fx, jacx)))

	Delta = 1.

	for it in range(maxiter):

		# Search direction
		p = cp.Variable(len(x))

		# Linearization of the objective function
		f_lin = fx + p.__rmatmul__(jacx)

		if norm == 1: obj = cp.norm1(f_lin)
		elif norm == 2: obj = cp.norm(f_lin)
		elif norm == np.inf: obj = cp.norm_inf(f_lin)
		elif norm == 'hinge': obj = cp.sum(cp.pos(f_lin))
		elif norm == None: obj = f_lin
		else: raise NotImplementedError
		# Now setup constraints
		nonlinear_constraints = []

		# First, constraints on "f"
		if obj_lb is not None:
			nonlinear_constraints.append(obj_lb <= f_lin)
		if obj_ub is not None:
			nonlinear_constraints.append(f_lin <= obj_ub)

		# Next, we add other nonlinear constraints
		for con, congrad, con_lb, con_ub in zip(constraints, constraint_grads, constraints_lb, constraints_ub):
			conx = con(x)
			congradx = congrad(x)
			#print "conx", conx, congradx
			if np.isfinite(con_lb):
				nonlinear_constraints.append(con_lb <= conx + p.__rmatmul__(congradx) )
			if np.isfinite(con_ub):
				nonlinear_constraints.append(conx + p.__rmatmul__(congradx) <= con_ub )

		# Constraints on the search direction specified by user
		search_step_constraints = search_constraints(x, p)

		# Append constraints from the domain of x
		domain_constraints = domain._build_constraints(x + p)

		stop = False
		for it2 in range(bt_maxiter):
			active_constraints = nonlinear_constraints + domain_constraints + search_step_constraints

			if it2 > 0:
				trust_region_constraints = [cp.norm(p) <= Delta]
				active_constraints += trust_region_constraints

			# Solve for the search direction

			with warnings.catch_warnings():
				warnings.simplefilter('ignore', PendingDeprecationWarning)
				try:
					problem = cp.Problem(cp.Minimize(obj), active_constraints)
					problem.solve(**kwargs)
					status = problem.status
				except cp.SolverError:
					if it2 == 0:
						status = 'unbounded'
					else:
						status = cp.SolverError

			if (status == 'unbounded' or status == 'unbounded_inaccurate') and it2 == 0:
				# On the first step, the trust region is off, allowing a potentially unbounded domain
				pass
			elif status in ['optimal', 'optimal_inaccurate']:
				# Otherwise, we've found a feasible step
				px = p.value
				# Evaluate new point along the trajectory
				x_new = trajectory(x, px, 1.)

				# Check for movement of the point
				if np.all(np.isclose(x, x_new, rtol = tol_dx, atol = 0)):
					stop = True
					break

				# Evaluate value at new point
				try:
					fx_new = np.array(f(x_new))
				except TypeError:
					fx_new = np.array([fi(x_new) for fi in f]).reshape(-1,)
				objval_new = objfun(fx_new)

				constraint_violation = 0.
				if obj_lb is not None:
					I = ~(obj_lb <= fx_new)
					constraint_violation += np.linalg.norm((fx_new - obj_lb)[I], 1)
				if obj_ub is not None:
					I = ~(fx_new <= obj_ub)
					constraint_violation += np.linalg.norm((fx_new - obj_ub)[I], 1)

				if objval_new < objval and np.isclose(constraint_violation, 0., rtol = 1e-10, atol = 1e-10):
					x = x_new
					fx = fx_new

					if np.abs(objval_new - objval) < tol_obj:
						stop = True
					objval = objval_new
					Delta = max(1., np.linalg.norm(px))
					break

				Delta *=0.5

			else:
				warnings.warn("Could not find acceptible step; stopping prematurely; %s" % (status,) )
				stop = True
				px = np.zeros(x.shape)

			#elif status in ['unbounded', 'unbounded_inaccurate']:
			#	raise UnboundedException
			#elirf status in ['infeasible']:
			#	raaise InfeasibleException
			#else:
			#	raise Exception(status)

		if it2 == bt_maxiter-1:
			stop = True

		# Update the jacobian information
		try:
			jacx = jac(x)
		except TypeError:
			jacx = np.array([jaci(x) for jaci in jac]).reshape(len(fx), len(x))

		if verbose:
			print('%4d | %+14.10e | %8.2e |  %8.2e | %8.2e |  %8.2e |'
				% (it+1, objval, np.linalg.norm(px), Delta, kkt_norm(fx, jacx), constraint_violation))
		if stop:
			break

	return x

#if __name__ == '__main__':
#	from polyridge import *
#
#	np.random.seed(3)
#	p = 3
#	m = 4
#	n = 1
#	M = 100
#
#	norm = np.inf
#	norm = 2
#	U = orth(np.random.randn(m,n))
#	coef = np.random.randn(len(LegendreTensorBasis(n,p)))
#	prf = PolynomialRidgeFunction(LegendreTensorBasis(n,p), coef, U)
#
#	X = np.random.randn(M,m)
#	fX = prf.eval(X)
#
#	pra = PolynomialRidgeApproximation(degree = p, subspace_dimension  = n, norm = norm, scale = True)
#
#	def residual(U_c):
#		r = pra._residual(X, fX, U_c)
#		return r
#
#	def jacobian(U_c):
#		U = U_c[:m*n].reshape(m,n)
#		pra.set_scale(X, U)
#		J = pra._jacobian(X, fX, U_c)
#		return J
#
#	# Trajectory
#	trajectory = lambda U_c, p, alpha: pra._trajectory(X, fX, U_c, p, alpha)
#
#	def search_constraints(U_c, pU_pc):
#	#	M, m = X.shape
#	#	N = len(self.basis)
#	#	n = self.subspace_dimension
#		U = U_c[:m*n].reshape(m,n)
#		constraints = [ pU_pc[k*m:(k+1)*m].__rmatmul__(U.T) == np.zeros(n) for k in range(n)]
#		return constraints
#
#
#	U0 = orth(np.random.randn(m,n))
#	U0 = U
#	c = np.random.randn(len(coef))
#	pra.set_scale(X, U0)
#	U_c0 = np.hstack([U0.flatten(), c])
#
#	U_c = sequential_lp(residual, U_c0, jacobian, search_constraints, norm = norm,
#		trajectory = trajectory, verbose = True)
#	print(U_c)
#	print(U)



__all__ = [
    "eval_form_pma",
    "ev_form_pma",
    "eval_form_ria",
    "ev_form_ria",
]

from grama import add_pipe, pipe, custom_formatwarning
from grama.eval_defaults import invariants_eval_model, invariants_eval_df
from .eval_defaults import eval_df
from numpy import array, argmin, ones, eye, zeros, sqrt, NaN, max
from numpy.linalg import norm as length
from numpy.random import multivariate_normal
from pandas import DataFrame, concat
from scipy.stats import norm
from scipy.optimize import minimize
from toolz import curry


## Utility Functions
# --------------------------------------------------
def make_T(model, df_corr):
    """Make covariance matrix from list

    Construct a covariance matrix from given entries. Check for correctness
    against a model.

    """
    ## TODO
    raise NotImplementedError


## FORM
# --------------------------------------------------
@curry
def eval_form_pma(
    model,
    betas=None,
    rels=None,
    pofs=None,
    cons=None,
    df_corr=None,
    df_det=None,
    append=True,
    tol=1e-3,
    n_maxiter=25,
    n_restart=1,
    verbose=False,
):
    r"""Tail quantile via FORM PMA

    Approximate the desired tail quantiles using the performance measure approach (PMA) of the first-order reliability method (FORM) [1]. Select limit states to minimize at desired quantile with one of: `betas` (reliability indices), `rels` (reliability values), or `pofs` (probabilities of failure). Provide confidence levels `cons` and estimator covariance `df_corr` to compute with margin in beta [2].

    Note that under the performance measure approach, the optimized limit state value `g` is sought to be non-negative $g \geq 0$. This is usually included as a constraint in optimization, which can be accomplished in by using ``gr.eval_form_pnd()` *within* a model definition---see the Examples below for more details.

    Args:
        model (gr.Model): Model to analyze
        df_det (DataFrame): Deterministic levels for evaluation; use "nom" for nominal deterministic levels.

        betas (dict): Target reliability indices;
            key   = limit state name; must be in model.out
            value = reliability index; beta = Phi^{-1}(reliability)
        rels (dict): Target reliability values;
            key   = limit state name; must be in model.out
            value = reliability
        pofs (dict): Target probabilities of failure (POFs);
            key   = limit state name; must be in model.out
            value = probability of failure; pof = 1 - reliability

        cons (dict or None): Target confidence levels;
            key   = limit state name; must be in model.out
            value = confidence level, \in (0, 1)
        df_corr (DataFrame or None): Sampling distribution covariance entries; parameters with no information assumed to be known exactly.

        n_maxiter (int): Maximum iterations for each optimization run
        n_restart (int): Number of restarts (== number of optimization runs)
        append (bool): Append MPP results for random values?
        verbose (bool): Print optimization results?

    Returns:
        DataFrame: Results of MPP search

    Notes:
        Since FORM PMA relies on optimization over the limit state, it is often beneficial to scale your limit state to keep values near unity.

    References:
        Tu, Choi, and Park, "A new study on reliability-based design optimization," Journal of Mechanical Design, 1999
        del Rosario, Fenrich, and Iaccarino, "Fast precision margin with the first-order reliability method," AIAA Journal, 2019

    Examples::

        import grama as gr
        from grama.models import make_cantilever_beam
        md_beam = make_cantilever_beam()
        ## Evaluate the reliability of specified designs
        (
            md_beam
            >> gr.ev_form_pma(
                # Specify target reliability
                betas=dict(g_stress=3, g_disp=3),
                # Analyze three different thicknesses
                df_det=gr.df_make(t=[2, 3, 4], w=3)
            )
        )

        ## Specify reliability in POF form
        (
            md_beam
            >> gr.ev_form_pma(
                # Specify target reliability
                pofs=dict(g_stress=1e-3, g_disp=1e-3),
                # Analyze three different thicknesses
                df_det=gr.df_make(t=[2, 3, 4], w=3)
            )
        )

        ## Build a nested model for optimization under uncertainty
        md_opt = (
            gr.Model("Beam Optimization")
            >> gr.cp_vec_function(
                fun=lambda df: gr.df_make(c_area=df.w * df.t),
                var=["w", "t"],
                out=["c_area"],
                name="Area objective",
            )
            >> gr.cp_vec_function(
                fun=lambda df: gr.eval_form_pma(
                    md_beam,
                    betas=dict(g_stress=3, g_disp=3),
                    df_det=df,
                    append=False,
                )
                var=["w", "t"],
                out=["g_stress", "g_disp"],
                name="Reliability constraints",
            )
            >> gr.cp_bounds(w=(2, 4), t=(2, 4))
        )
        # Run the optimization
        (
            md_opt
            >> gr.ev_min(
                out_min="c_area",
                out_geq=["g_stress", "g_disp"],
            )
        )

    """
    ## Check invariants
    invariants_eval_model(model)
    invariants_eval_df(df_corr, arg_name="df_corr", acc_none=True)
    invariants_eval_df(df_det, arg_name="df_det", valid_str=["nom"])
    # Check that reliability targets given by one argument only
    if all((betas is None, rels is None, pofs is None)):
        raise ValueError(
            "Must provide reliability targets as keyword argument `betas`, `rels`, OR `pofs`"
        )
    if sum((betas is not None, rels is not None, pofs is not None)) > 1:
        raise ValueError(
            "Must provide *only* one of `betas`, `rels`, OR `pofs`"
        )
    # Convert reliability targets to `betas`
    if rels is not None:
        betas = {k: norm.ppf(v) for (k, v) in rels.items()}
    if pofs is not None:
        betas = {k: norm.ppf(1 - v) for (k, v) in pofs.items()}

    # Check `betas` invariants
    if not set(betas.keys()).issubset(set(model.out)):
        raise ValueError("betas.keys() must be subset of model.out")
    if not cons is None:
        if not (set(cons.keys()) == set(betas.keys())):
            raise ValueError("cons.keys() must be same as betas.keys()")
        else:
            if df_corr is None:
                raise ValueError("Must provide df_corr is using cons")
        raise NotImplementedError

    df_det = model.var_outer(
        DataFrame(data=zeros((1, model.n_var_rand)), columns=model.var_rand),
        df_det=df_det,
    )
    df_det = df_det[model.var_det]

    df_return = DataFrame()
    for ind in range(df_det.shape[0]):
        ## Loop over objectives
        for key in betas.keys():
            ## Temp dataframe
            df_inner = df_det.iloc[[ind]].reset_index(drop=True)

            ## Construct lambdas
            def objective(z):
                ## Transform: standard normal-to-random variable
                df_norm = DataFrame(data=[z], columns=model.var_rand)
                df_rand = model.norm2rand(df_norm)
                df = model.var_outer(df_rand, df_det=df_inner)

                df_res = eval_df(model, df=df)
                g = df_res[key].iloc[0]

                # return (g, jac)
                return g

            def con_beta(z):
                return z.dot(z) - (betas[key]) ** 2

            ## Use conservative direction for initial guess
            signs = array([model.density.marginals[k].sign for k in model.var_rand])
            if length(signs) > 0:
                z0 = betas[key] * signs / length(signs)
            else:
                z0 = betas[key] * ones(model.n_var_rand) / sqrt(model.n_var_rand)

            ## Minimize
            res_all = []
            for jnd in range(n_restart):
                res = minimize(
                    objective,
                    z0,
                    args=(),
                    method="SLSQP",
                    jac=False,
                    tol=tol,
                    options={"maxiter": n_maxiter, "disp": False},
                    constraints=[{"type": "eq", "fun": con_beta}],
                )
                # Append only a successful result
                if res["status"] == 0:
                    res_all.append(res)
                # Set a random start; repeat
                z0 = multivariate_normal([0] * model.n_var_rand, eye(model.n_var_rand))
                z0 = z0 / length(z0) * betas[key]

            # Choose value among restarts
            n_iter_total = sum([res_all[i].nit for i in range(len(res_all))])
            if len(res_all) > 0:
                i_star = argmin([res.fun for res in res_all])
                x_star = res_all[i_star].x
                fun_star = res_all[i_star].fun
                if verbose:
                    print("out = {}: Optimization successful".format(key))
                    print("n_iter = {}".format(res_all[i_star].nit))
                    print("n_iter_total = {}".format(n_iter_total))
            else:
                ## WARNING
                x_star = [NaN] * model.n_var_rand
                fun_star = NaN
                if verbose:
                    print("out = {}: Optimization unsuccessful".format(key))
                    print("n_iter = {}".format(res_all[i_star].nit))
                    print("n_iter_total = {}".format(n_iter_total))

            ## Extract results
            if append:
                df_inner = concat(
                    (
                        df_inner,
                        model.norm2rand(
                            DataFrame(data=[x_star], columns=model.var_rand)
                        ),
                    ),
                    axis=1,
                    sort=False,
                )
            df_inner[key] = [fun_star]
            df_return = concat((df_return, df_inner), axis=0, sort=False)

    if not append:
        df_return = (
            df_return.groupby(model.var_det)
            .agg({s: max for s in betas.keys()})
            .reset_index()
        )

    return df_return


ev_form_pma = add_pipe(eval_form_pma)


@curry
def eval_form_ria(
    model,
    limits=None,
    format="betas",
    cons=None,
    df_corr=None,
    df_det=None,
    append=True,
    tol=1e-3,
    n_maxiter=25,
    n_restart=1,
    verbose=False,
):
    r"""Tail reliability via FORM RIA

    Approximate the desired tail probability using the reliability index approach (RIA) of the first-order reliability method (FORM) [1]. Select limit states to analyze with list input `limits`. Choose output type using `format` argument (`betas` for reliability indices, `rels` for realibility values, `pofs` for probabilify of failure values). Provide confidence levels `cons` and estimator covariance `df_corr` to compute with margin in beta [2].

    Note that the reliability index approach (RIA) is generally less stable than the performance measure approach (PMA). Consider using ``gr.eval_form_pma()`` instead, particularly when using FORM to optimize a design.

    Args:
        model (gr.Model): Model to analyze
        limits (list): Target limit states; must be in model.out; limit state assumed to be critical at g == 0.
        format (str): One of ("betas", "rels", "pofs"). Format for computed reliability information

        cons (dict or None): Target confidence levels;
            key   = limit state name; must be in model.out
            value = confidence level, \in (0, 1)
        df_corr (DataFrame or None): Sampling distribution covariance entries; parameters with no information assumed to be known exactly.
        df_det (DataFrame): Deterministic levels for evaluation; use "nom" for nominal deterministic levels.
        n_maxiter (int): Maximum iterations for each optimization run
        n_restart (int): Number of restarts (== number of optimization runs)
        append (bool): Append MPP results for random values?
        verbose (bool): Print optimization results?

    Returns:
        DataFrame: Results of MPP search

    Notes:
        Since FORM RIA relies on optimization over the limit state, it is often beneficial to scale your limit state to keep values near unity.

    References:
        [1] Tu, Choi, and Park, "A new study on reliability-based design optimization," Journal of Mechanical Design, 1999
        [2] del Rosario, Fenrich, and Iaccarino, "Fast precision margin with the first-order reliability method," AIAA Journal, 2019

    Examples::

        import grama as gr
        from grama.models import make_cantilever_beam
        md_beam = make_cantilever_beam()
        ## Evaluate the reliability of specified designs
        (
            md_beam
            >> gr.ev_form_ria(
                # Specify limit states to analyze
                limits=("g_stress", "g_disp"),
                # Analyze three different thicknesses
                df_det=gr.df_make(t=[2, 3, 4], w=3)
            )
        )

    """
    ## Check invariants
    invariants_eval_model(model)
    invariants_eval_df(df_corr, arg_name="df_corr", acc_none=True)
    invariants_eval_df(df_det, arg_name="df_det", valid_str=["nom"])
    if limits is None:
        raise ValueError(
            "Must provide `limits` keyword argument to define reliability targets"
        )
    if not set(limits).issubset(set(model.out)):
        raise ValueError("`limits` must be subset of model.out")
    if not cons is None:
        if not (set(cons.keys()) == set(limits)):
            raise ValueError("cons.keys() must be same as limits")
        else:
            if df_corr is None:
                raise ValueError("Must provide df_corr is using cons")
        raise NotImplementedError

    df_det = model.var_outer(
        DataFrame(data=zeros((1, model.n_var_rand)), columns=model.var_rand),
        df_det=df_det,
    )
    df_det = df_det[model.var_det]

    # df_return = DataFrame(columns=model.var_rand + model.var_det + limits)
    df_return = DataFrame()
    for ind in range(df_det.shape[0]):
        ## Loop over objectives
        for key in limits:
            ## Temp dataframe
            df_inner = df_det.iloc[[ind]].reset_index(drop=True)

            ## Construct lambdas
            def fun_jac(z):
                ## Squared reliability index
                fun = z.dot(z)
                jac = 2 * z * length(z)

                return (fun, jac)

            def con_limit(z):
                ## Transform: standard normal-to-random variable
                df_norm = DataFrame(data=[z], columns=model.var_rand)
                df_rand = model.norm2rand(df_norm)
                df = model.var_outer(df_rand, df_det=df_inner)

                ## Eval limit state
                df_res = eval_df(model, df=df)
                g = df_res[key].iloc[0]

                return g

            ## Use conservative direction for initial guess
            signs = array([model.density.marginals[k].sign for k in model.var_rand])
            if length(signs) > 0:
                z0 = signs / length(signs)
            else:
                z0 = ones(model.n_var_rand) / sqrt(model.n_var_rand)

            ## Minimize
            res_all = []
            for jnd in range(n_restart):
                res = minimize(
                    fun_jac,
                    z0,
                    args=(),
                    method="SLSQP",
                    jac=True,
                    tol=tol,
                    options={"maxiter": n_maxiter, "disp": False},
                    constraints=[{"type": "eq", "fun": con_limit}],
                )
                # Append only a successful result
                if res["status"] == 0:
                    res_all.append(res)
                # Set a random start; repeat
                z0 = multivariate_normal([0] * model.n_var_rand, eye(model.n_var_rand))
                z0 = z0 / length(z0)

            # Choose value among restarts
            n_iter_total = sum([res_all[i].nit for i in range(len(res_all))])
            if len(res_all) > 0:
                i_star = argmin([res.fun for res in res_all])
                x_star = res_all[i_star].x
                fun_star = sqrt(res_all[i_star].fun)
                if verbose:
                    print("out = {}: Optimization successful".format(key))
                    print("n_iter = {}".format(res_all[i_star].nit))
                    print("n_iter_total = {}".format(n_iter_total))
            else:
                ## WARNING
                x_star = [NaN] * model.n_var_rand
                fun_star = NaN
                if verbose:
                    print("out = {}: Optimization unsuccessful".format(key))
                    print("n_iter = {}".format(res_all[i_star].nit))
                    print("n_iter_total = {}".format(n_iter_total))

            ## Extract results
            if append:
                df_inner = concat(
                    (
                        df_inner,
                        model.norm2rand(
                            DataFrame(data=[x_star], columns=model.var_rand)
                        ),
                    ),
                    axis=1,
                    sort=False,
                )
            df_inner["beta_" + key] = [fun_star]
            df_return = concat((df_return, df_inner), axis=0, sort=False)

    ## Collapse beta values w/o MPP results
    beta_names = ["beta_" + s for s in limits]
    if not append:
        df_return = (
            df_return.groupby(model.var_det) \
                     .agg({n: max for n in beta_names}).reset_index()
        )

    ## Convert betas if other format requested
    if format == "rels":
        df_return = df_return.apply(lambda col: norm.cdf(col) if col.name in beta_names else col) \
                             .rename(columns={"beta_" + s: "rel_" + s for s in limits})

    elif format == "pofs":
        df_return = df_return.apply(lambda col: 1 - norm.cdf(col) if col.name in beta_names else col) \
                             .rename(columns={"beta_" + s: "pof_" + s for s in limits})

    elif not (format == "betas"):
        print("    format = '{}' unrecognized; returning reliability indices".format(format))

    return df_return


ev_form_ria = add_pipe(eval_form_ria)




class EmptyDomainException(Exception):
	r""" Raised when trying to call an operation that does not work on an empty domain
	"""
	pass

class UnboundedDomainException(Exception):
	r""" Raised when trying to preform an operation on a domain that is ill-posed on an unbounded domain
	"""
	pass

class SolverError(Exception):
	r""" A problem with the solver
	"""
	pass	

class UnderdeterminedException(Exception):
	pass



__all__ = [
    "binomial_ci",
    "corr",
    "mean",
    "mean_lo",
    "mean_up",
    "IQR",
    "quant",
    "pint_lo",
    "pint_lo_index",
    "pint_up",
    "pint_up_index",
    "pr",
    "pr_lo",
    "pr_up",
    "var",
    "sd",
    "skew",
    "kurt",

    "min",
    "max",
    "sum",
    "median",

    "first",
    "last",

    "n",
    "nth",
    "n_distinct",
    "neff_is",

    "mad",
    "mead",
    "mse",
    "rmse",
    "ndme",
    "rsq",
]

from .base import make_symbolic, Intention
from .vector import order_series_by
from numpy import array, sqrt, power, nan, isnan
from scipy.stats import norm, pearsonr, spearmanr, kurtosis, nhypergeom
from scipy.stats import skew as spskew


# ------------------------------------------------------------------------------
# Series summary functions
# ------------------------------------------------------------------------------


@make_symbolic
def mean(series):
    """
    Returns the mean of a series.

    Args:
        series (pandas.Series): column to summarize.
    """

    # if np.issubdtype(series.dtype, np.number):
    #     return series.mean()
    # else:
    #     return np.nan
    return series.mean()


# Mean CI helpers
@make_symbolic
def mean_lo(series, alpha=0.005):
    """Return a confidence interval (lower bound) for the mean

    Uses a central limit approximation for a lower confidence bound of an estimated mean. That is:

        m - q(alpha) * s / sqrt(n)

    where

        m = sample mean
        s = sample standard deviation
        n = sample size
        q(alpha) = alpha-level lower-quantile of standard normal
                 = (-norm.ppf(alpha))

    For a two-sided interval at a confidence level of ``C``, set ``alpha = 1 - C`` and use ``[gr.mean_lo(X, alpha=alpha/2), gr.mean_up(X, alpha=alpha/2)``. Note that the default ``alpha`` level for both helpers is calibrated for a two-sided interval with ``C = 0.99``.

    Args:
        series (pandas.Series): column to summarize
        alpha (float): alpha-level for calculation. Note that the confidence level C is given by ``C = 1 - alpha``.

    Returns:
        float: Lower confidence interval for the mean
    """
    m = mean(series)
    s = sd(series)
    n_sample = len(series)

    return m - (-norm.ppf(alpha)) * s / sqrt(n_sample)


@make_symbolic
def mean_up(series, alpha=0.005):
    """Return a confidence interval (upper bound) for the mean

    Uses a central limit approximation for a upper confidence bound of an estimated mean. That is:

        m + q(alpha) * s / sqrt(n)

    where

        m = sample mean
        s = sample standard deviation
        n = sample size
        q(alpha) = alpha-level lower-quantile of standard normal
                 = (-norm.ppf(alpha))

    For a two-sided interval at a confidence level of ``C``, set ``alpha = 1 - C`` and use ``[gr.mean_lo(X, alpha=alpha/2), gr.mean_up(X, alpha=alpha/2)``. Note that the default ``alpha`` level for both helpers is calibrated for a two-sided interval with ``C = 0.99``.

    Args:
        series (pandas.Series): column to summarize
        alpha (float): alpha-level for calculation. Note that the confidence level C is given by ``C = 1 - alpha``.

    Returns:
        float: Upper confidence interval for the mean
    """
    m = mean(series)
    s = sd(series)
    n_sample = len(series)

    return m + (-norm.ppf(alpha)) * s / sqrt(n_sample)


@make_symbolic
def skew(series, bias=True, nan_policy="propagate"):
    """Returns the skewness of a series.

    Args:
        series (pandas.Series): column to summarize.
        bias (bool): Correct for bias?
        nan_policy (str): How to handle NaN values:
            - "propagate": return NaN
            - "raise": throws an error
            - "omit": remove NaN before calculating skew

    """

    return spskew(series, bias=bias, nan_policy="propagate")


@make_symbolic
def kurt(series, bias=True, nan_policy="propagate", excess=False):
    """Returns the kurtosis of a series.

    A distribution with kurtosis greater than three is called *leptokurtic*; such a distribution has "fatter" tails and will tend to exhibit more outliers. A distribution with kurtosis less than three is called *platykurtic*; such a distribution has less-fat tails and will tend to exhibit fewer outliers.

    Args:
        series (pandas.Series): column to summarize.
        bias (bool): Correct for bias?
        excess (bool): Return excess kurtosis (excess = kurtosis - 3).
            Note that a normal distribution has kurtosis == 3, which
            informs the excess kurtosis definition.
        nan_policy (str): How to handle NaN values:
            - "propagate": return NaN
            - "raise": throws an error
            - "omit": remove NaN before calculating skew

    """

    return kurtosis(series, fisher=excess, bias=bias, nan_policy="propagate")

@make_symbolic
def first(series, order_by=None):
    """
    Returns the first value of a series.

    Args:
        series (pandas.Series): column to summarize.

    Kwargs:
        order_by: a pandas.Series or list of series (can be symbolic) to order
            the input series by before summarization.
    """

    if order_by is not None:
        series = order_series_by(series, order_by)
    first_s = series.iloc[0]
    return first_s


@make_symbolic
def last(series, order_by=None):
    """
    Returns the last value of a series.

    Args:
        series (pandas.Series): column to summarize.

    Kwargs:
        order_by: a pandas.Series or list of series (can be symbolic) to order the input series by before summarization.
    """

    if order_by is not None:
        series = order_series_by(series, order_by)
    last_s = series.iloc[series.size - 1]
    return last_s


@make_symbolic
def nth(series, n, order_by=None):
    """
    Returns the nth value of a series.

    Args:
        series (pandas.Series): column to summarize.
        n (integer): position of desired value. Returns `NaN` if out of range.

    Kwargs:
        order_by: a pandas.Series or list of series (can be symbolic) to order the input series by before summarization.
    """

    if order_by is not None:
        series = order_series_by(series, order_by)
    try:
        return series.iloc[n]
    except:
        return nan


@make_symbolic
def n(series=None):
    """
    Returns the length of a series.

    Args:
        series (pandas.Series): column to summarize. Default is the size of the parent DataFrame.

    Examples::

        import grama as gr
        from grama.data import df_diamonds
        DF = gr.Intention()

        ## Count entries in series
        gr.n(df_diamonds.cut)
        ## Use implicit mode to get size of current DataFrame
        (
            df_diamonds
            >> gr.tf_mutate(n_total=gr.n())
        )
        ## Use implicit mode in groups
        (
            df_diamonds
            >> gr.tf_group_by(DF.cut)
            >> gr.tf_mutate(n_cut=gr.n())
        )

    """
    # "Implicit mode": Default to current DataFrame
    if series is None:
        DF = Intention()
        series = DF.index

    n_s = series.size
    return n_s


@make_symbolic
def n_distinct(series):
    """
    Returns the number of distinct values in a series.

    Args:
        series (pandas.Series): column to summarize.
    """

    n_distinct_s = series.unique().size
    return n_distinct_s


@make_symbolic
def IQR(series):
    """
    Returns the inter-quartile range (IQR) of a series.

    The IRQ is defined as the 75th quantile minus the 25th quantile values.

    Args:
        series (pandas.Series): column to summarize.
    """

    iqr_s = series.quantile(0.75) - series.quantile(0.25)
    return iqr_s


@make_symbolic
def quant(series, p=None):
    """
    Returns the specified quantile value.

    Args:
        series (pandas.Series): Column to summarize
        p (float): Fraction for desired quantile, 0 <= p <= 1

    Returns:
        float: Desired quantile
    """

    return series.quantile(p)


@make_symbolic
def min(series):
    """
    Returns the minimum value of a series.

    Args:
        series (pandas.Series): column to summarize.
    """

    min_s = series.min()
    return min_s


@make_symbolic
def max(series):
    """
    Returns the maximum value of a series.

    Args:
        series (pandas.Series): column to summarize.
    """

    max_s = series.max()
    return max_s


@make_symbolic
def median(series):
    """
    Returns the median value of a series.

    Args:
        series (pandas.Series): column to summarize.
    """

    # if np.issubdtype(series.dtype, np.number):
    #     return series.median()
    # else:
    #     return np.nan
    return series.median()


@make_symbolic
def var(series):
    """
    Returns the variance of values in a series.

    Args:
        series (pandas.Series): column to summarize.
    """
    # if np.issubdtype(series.dtype, np.number):
    #     return series.var()
    # else:
    #     return np.nan
    return series.var()


@make_symbolic
def sd(series):
    """
    Returns the standard deviation of values in a series.

    Args:
        series (pandas.Series): column to summarize.
    """

    # if np.issubdtype(series.dtype, np.number):
    #     return series.std()
    # else:
    #     return np.nan
    return series.std()


@make_symbolic
def sum(series):
    """
    Returns the sum of values in a series.

    Args:
        series (pandas.Series): column to summarize.
    """

    return series.sum()


@make_symbolic
def mad(series_pred, series_meas):
    """Compute MAD

    Returns the mean absolute deviation (MAD) between predicted and measured
    values.

    Args:
        series_pred (pandas.Series): column of predicted values
        series_meas (pandas.Series): column of measured values

    Returns:
        float: Mean absolute deviation (MAD)
    """

    return (series_pred - series_meas).abs().mean()


@make_symbolic
def mead(series_pred, series_meas):
    """Compute median absolute deviation

    Returns the median absolute deviation (MEAD) between predicted and measured
    values.

    Args:
        series_pred (pandas.Series): column of predicted values
        series_meas (pandas.Series): column of measured values

    Returns:
        float: Median absolute deviation (MEAD)
    """

    return (series_pred - series_meas).abs().median()


@make_symbolic
def mse(series_pred, series_meas):
    """Compute MSE

    Returns the mean-square-error (MSE) between predicted and measured
    values.

    Args:
        series_pred (pandas.Series): column of predicted values
        series_meas (pandas.Series): column of measured values

    Returns:
        float: Mean squared error (MSE)
    """

    return (series_pred - series_meas).pow(2).mean()


@make_symbolic
def rmse(series_pred, series_meas):
    """Compute RMSE

    Returns the root-mean-square-error (RMSE) between predicted and measured
    values.

    Args:
        series_pred (pandas.Series): column of predicted values
        series_meas (pandas.Series): column of measured values

    Returns:
        float: Root-mean squared error (RMSE)

    """

    return sqrt((series_pred - series_meas).pow(2).mean())


@make_symbolic
def rel_mse(series_pred, series_meas):
    """Compute relative MSE

    Returns the "relative" mean-square-error (MSE) between predicted and measured values. All differences between predictions and measured values are normalized by the measured value prior to taking the MSE.

    Args:
        series_pred (pandas.Series): column of predicted values
        series_meas (pandas.Series): column of measured values

    Returns:
        float: Mean Squared Error (MSE)

    """

    return ((series_pred - series_meas) / series_meas).pow(2).mean()


@make_symbolic
def rsq(series_pred, series_meas):
    """Compute coefficient of determination

    Returns the coefficient of determination (aka R^2) between predicted and measured values. Theoretically 0 <= R^2 <= 1, with R^2 == 0 corresponding to a model no more predictive than guessing the mean of observed values, and R^2 == 1 corresponding to a perfect model. Note that sampling variability can lead to R^2 < 0 or R^2 > 1.

    Args:
        series_pred (pandas.Series): column of predicted values
        series_meas (pandas.Series): column of measured values

    """

    y_mean = series_meas.mean()

    SS_res = power(series_meas - series_pred, 2).sum()
    SS_tot = power(series_meas - y_mean, 2).sum()

    return 1 - SS_res / SS_tot


@make_symbolic
def ndme(series_pred, series_meas):
    """Compute non-dimensional model error

    Returns the non-dimensional model error (NDME) between predicted and measured values. The NDME is related to the coefficient of determination (aka R^2) via

        NDME = sqrt(1 - R^2)

    Args:
        series_pred (pandas.Series): column of predicted values
        series_meas (pandas.Series): column of measured values

    """

    return sqrt(1 - rsq(series_pred, series_meas))


@make_symbolic
def binomial_ci(series, alpha=0.05, side="both"):
    """Returns a binomial confidence interval

    Computes a binomial confidence interval based on boolean data. Uses Wilson interval

    Args:
        series (pandas.Series): Column to summarize; must be boolean or 0/1.
        alpha (float): Confidence level; value in (0, 1)
        side (string): Chosen side of interval
            - "both": Return a 2-tuple of series
            - "lo": Return the lower interval bound
            - "up": Return the upper interval bound
    """
    n_s = series.sum()
    n_t = len(series)
    n_f = n_t - n_s
    z = -norm.ppf(alpha / 2)

    mid = (n_s + 0.5 * z ** 2) / (n_t + z ** 2)
    delta = z / (n_t + z ** 2) * sqrt(n_s * n_f / n_t + 0.25 * z ** 2)
    lo = mid - delta
    up = mid + delta

    if side == "both":
        return (lo, up)
    if side == "lo":
        return lo
    if side == "up":
        return up
    else:
        raise ValueError("side value {} not recognized".format(side))

# Probability helpers
# --------------------------------------------------
@make_symbolic
def pr(series):
    """Estimate a probability

    Estimate a probability from a random sample. Provided series must be boolean, with 1 corresponding to the event of interest.

    Use logical statements together with column values to construct a boolean indicator for the event you're interested in. Remember that you can chain multiple statements with logical and `&` and or `|` operators. See the examples below for more details.

    Args:
        series (pandas.Series): Column to summarize; must be boolean or 0/1.

    Examples::

        import grama as gr
        DF = gr.Intention()
        ## Cantilever beam examples
        from grama.models import make_cantilever_beam
        md_beam = make_cantilever_beam()

        ## Estimate probabilities
        (
            md_beam
            # Generate large
            >> gr.ev_sample(n=1e5, df_det="nom")
            # Estimate probabilities of failure
            >> gr.tf_summarize(
                pof_stress=gr.pr(DF.g_stress <= 0),
                pof_disp=gr.pr(DF.g_disp <= 0),
                pof_joint=gr.pr( (DF.g_stress <= 0) & (DF.g_disp) ),
                pof_either=gr.pr( (DF.g_stress <= 0) | (DF.g_disp) ),
            )
        )

    """
    return series.mean()


@make_symbolic
def pr_lo(series, alpha=0.005):
    r"""Estimate a confidence interval for a probability

    Estimate the lower side of a confidence interval for a probability from a random sample. Provided series must be boolean, with 1 corresponding to the event of interest.

    Uses Wilson interval method.

    Use logical statements together with column values to construct a boolean indicator for the event you're interested in. Remember that you can chain multiple statements with logical and `&` and or `|` operators. See the documentation for `gr.pr()` for more details and examples.

    For a two-sided interval at a confidence level of ``C``, set ``alpha = 1 - C`` and use ``[gr.mean_lo(X, alpha=alpha/2), gr.mean_up(X, alpha=alpha/2)``. Note that the default ``alpha`` level for both helpers is calibrated for a two-sided interval with ``C = 0.99``.

    Args:
        series (pandas.Series): Column to summarize; must be boolean or 0/1.
        alpha (float): alpha-level for calculation, in (0, 1)
            Note that the confidence level C is given by C = 1 - alpha

    Returns:
        float: Lower confidence interval

    Examples::

        import grama as gr
        DF = gr.Intention()
        ## Cantilever beam examples
        from grama.models import make_cantilever_beam
        md_beam = make_cantilever_beam()

        ## Estimate probabilities
        (
            md_beam
            # Generate large
            >> gr.ev_sample(n=1e5, df_det="nom")
            # Estimate probabilities with a confidence interval
            >> gr.tf_summarize(
                pof_lo=gr.pr_lo(DF.g_stress <= 0),
                pof=gr.pr(DF.g_stress <= 0),
                pof_up=gr.pr_up(DF.g_stress <= 0),
            )
        )
    """
    up = binomial_ci(series, alpha=alpha, side="lo")
    return up


@make_symbolic
def pr_up(series, alpha=0.005):
    r"""

    Estimate the upper side of a confidence interval for a probability from a random sample. Provided series must be boolean, with 1 corresponding to the event of interest.

    Uses Wilson interval method.

    Use logical statements together with column values to construct a boolean indicator for the event you're interested in. Remember that you can chain multiple statements with logical and `&` and or `|` operators. See the documentation for `gr.pr()` for more details and examples.

    For a two-sided interval at a confidence level of ``C``, set ``alpha = 1 - C`` and use ``[gr.mean_lo(X, alpha=alpha/2), gr.mean_up(X, alpha=alpha/2)``. Note that the default ``alpha`` level for both helpers is calibrated for a two-sided interval with ``C = 0.99``.

    Args:
        series (pandas.Series): Column to summarize; must be boolean or 0/1.
        alpha (float): alpha-level for calculation, in (0, 1)
            Note that the confidence level C is given by C = 1 - alpha

    Returns:
        float: Upper confidence interval

    Examples::

        import grama as gr
        DF = gr.Intention()
        ## Cantilever beam examples
        from grama.models import make_cantilever_beam
        md_beam = make_cantilever_beam()

        ## Estimate probabilities
        (
            md_beam
            # Generate large
            >> gr.ev_sample(n=1e5, df_det="nom")
            # Estimate probabilities with a confidence interval
            >> gr.tf_summarize(
                pof_lo=gr.pr_lo(DF.g_stress <= 0),
                pof=gr.pr(DF.g_stress <= 0),
                pof_up=gr.pr_up(DF.g_stress <= 0),
            )
        )
    """
    up = binomial_ci(series, alpha=alpha, side="up")
    return up


@make_symbolic
def corr(series1, series2, method="pearson", res="corr", nan_drop=False):
    r"""Computes a correlation coefficient

    Computes a correlation coefficient using either the pearson or spearman formulation.

    Args:
        series1 (pandas.Series): Column 1 to study
        series2 (pandas.Series): Column 2 to study
        method (str): Method to use; either "pearson" or "spearman"
        res (str): Quantities to return; either "corr" or "both"
        na_drop (bool): Drop NaN values before computation?

    Returns:
        pandas.Series: correlation coefficient

    """
    if nan_drop:
        ids = (isnan(series1) | isnan(series2))
    else:
        ids = array([False] * len(series1))

    if method == "pearson":
        r, p = pearsonr(series1[~ids], series2[~ids])
    elif method == "spearman":
        r, p = spearmanr(series1[~ids], b=series2[~ids])
    else:
        raise ValueError("method {} not supported".format(method))

    if res == "corr":
        return r
    if res == "both":
        return r, p
    else:
        raise ValueError("res {} not supported".format(res))


# Distribution-free Prediction Intervals
# --------------------------------------------------
def pint_lo_index(n, m, j, alpha):
    r"""PI lower bound index

    Compute the order statistic index for the lower bound of a distribution-free
    prediction interval.

    """
    l = int(n - nhypergeom.ppf(1 - alpha, m + n, n, m - j  + 1))
    if l <= 0:
        raise ValueError(
            "Insufficient series length for requested `alpha` level; " +
            "obtain more observations or choose a larger value for `alpha`."
        )
    return l


def pint_up_index(n, m, j, alpha):
    r"""PI upper bound index

    Compute the order statistic index for the upper bound of a distribution-free
    prediction interval.

    """
    u = int(nhypergeom.ppf(1 - alpha, m + n, n, j) + 1)
    if u >= n + 1:
        raise ValueError(
            "Insufficient series length for requested `alpha` level; " +
            "obtain more observations or choose a larger value for `alpha`."
        )
    return u


@make_symbolic
def pint_lo(series, m=1, j=1, alpha=0.005):
    r"""Lower prediction interval

    Compute a one-sided lower prediction interval using a distribution-free approach.

    For a two-sided interval at a confidence level of ``C``, set ``alpha = 1 - C`` and use ``[gr.pint_lo(X, alpha=alpha/2), gr.pint_up(X, alpha=alpha/2)``. Note that the default ``alpha`` level for both helpers is calibrated for a two-sided interval with ``C = 0.99``.

    Args:
        series (pd.Series): Dataset to analyze

    Kwargs:
        m (int): Number of observations in future dataset (Default m=1)
        j (int): Order statistic to target; 1 <= j <= m (Default j=1)
        alpha (float): alpha-level for calculation, in (0, 1). Note that the confidence level C is given by ``C = 1 - alpha``.

    References:
        Hahn, Gerald J., and William Q. Meeker. Statistical intervals: a guide for practitioners. Vol. 92. John Wiley & Sons, 2011.

    """
    n = len(series)
    l = pint_lo_index(n, m, j, alpha)
    return series.sort_values().iloc[l - 1] # for 0-based indexing

@make_symbolic
def pint_up(series, m=1, j=1, alpha=0.005):
    r"""Upper prediction interval

    Compute a one-sided upper prediction interval using a distribution-free approach.

    For a two-sided interval at a confidence level of ``C``, set ``alpha = 1 - C`` and use ``[gr.pint_lo(X, alpha=alpha/2), gr.pint_up(X, alpha=alpha/2)``. Note that the default ``alpha`` level for both helpers is calibrated for a two-sided interval with ``C = 0.99``.

    Args:
        series (pd.Series): Dataset to analyze

    Kwargs:
        m (int): Number of observations in future dataset (Default m=1)
        j (int): Order statistic to target; 1 <= j <= m (Default j=1)
        alpha (float): alpha-level for calculation, in (0, 1). Note that the confidence level C is given by ``C = 1 - alpha``.

    References:
        Hahn, Gerald J., and William Q. Meeker. Statistical intervals: a guide for practitioners. Vol. 92. John Wiley & Sons, 2011.

    """
    n = len(series)
    u = pint_up_index(n, m, j, alpha)
    return series.sort_values().iloc[u - 1] # for 0-based indexing

# ------------------------------------------------------------------------------
# Effective Sample Size helpers
# ------------------------------------------------------------------------------
@make_symbolic
def neff_is(series):
    """Importance sampling n_eff

    Computes the effective sample size based on importance sampling weights. See Equation 9.13 of Owen (2013) for details. See ``gr.tran_reweight()`` for more details.

    Args:
        series (pandas.Series): column of importance sampling weights.

    References:
        A.B. Owen, "Monte Carlo theory, methods and examples" (2013)

    """

    w = series.mean()
    w2 = (series**2).mean()
    return len(series) * w**2 / w2



# 2018 (c) Jeffrey M. Hokanson and Caleb Magruder
from __future__ import print_function, division

import warnings
import numpy as np
import scipy.linalg
import scipy as sp


__all__ = ['linesearch_armijo',
	'gauss_newton',
	'trajectory_linear',
	]


class BadStep(Exception):
	pass


def trajectory_linear(x0, p, t):
	return x0 + t * p


def linesearch_armijo(f, g, p, x0, bt_factor=0.5, ftol=1e-4, maxiter=40, trajectory = trajectory_linear, fx0 = None):
	"""Back-Tracking Line Search to satify Armijo Condition

		f(x0 + alpha*p) < f(x0) + alpha * ftol * <g,p>

	Parameters
	----------
	f : callable
		objective function, f: R^n -> R
	g : np.array((n,))
		gradient
	p : np.array((n,))
		descent direction
	x0 : np.array((n,))
		current location
	bt_factor : float [optional] default = 0.5
		backtracking factor
	ftol : float [optional] default = 1e-4
		coefficient in (0,1); see Armijo description in Nocedal & Wright
	maxiter : int [optional] default = 10
		maximum number of iterations of backtrack
	trajectory: function(x0, p, t) [Optional]
		Function that returns next iterate 
	Returns
	-------
	float
		alpha: backtracking coefficient (alpha = 1 implies no backtracking)
	"""
	
	dg = np.inner(g, p)
	assert dg <= 0, 'Descent direction p is not a descent direction: p^T g = %g >= 0' % (dg, )

	alpha = 1

	if fx0 is None:
		fx0 = f(x0)

	fx0_norm = np.linalg.norm(fx0)
	x = np.copy(x0)
	fx = np.inf
	success = False
	for it in range(maxiter):
		try:
			x = trajectory(x0, p, alpha)
			fx = f(x)
			fx_norm = np.linalg.norm(fx)
			if fx_norm < fx0_norm + alpha * ftol * dg:
				success = True
				break
		except BadStep:
			pass
			
		alpha *= bt_factor

	# If we haven't found a good step, stop
	if not success:
		alpha = 0
		x = x0
		fx = fx0
	return x, alpha, fx


def gauss_newton(f, F, x0, tol=1e-10, tol_normdx=1e-12, 
	maxiter=100, linesearch=None, verbose=0, trajectory=None, gnsolver = None):
	r"""A Gauss-Newton solver for unconstrained nonlinear least squares problems.

	Given a vector valued function :math:`\mathbf{f}:\mathbb{R}^m \to \mathbb{R}^M`
	and its Jacobian :math:`\mathbf{F}:\mathbb{R}^m\to \mathbb{R}^{M\times m}`,
	solve the nonlinear least squares problem:

	.. math::

		\min_{\mathbf{x}\in \mathbb{R}^m} \| \mathbf{f}(\mathbf{x})\|_2^2.

	Normal Gauss-Newton computes a search direction :math:`\mathbf{p}\in \mathbb{R}^m`
	at each iterate by solving least squares problem

	.. math::
		
		\mathbf{p}_k \leftarrow \mathbf{F}(\mathbf{x}_k)^+ \mathbf{f}(\mathbf{x}_k)

	and then computes a new step by solving a line search problem for a step length :math:`\alpha`
	satisfying the Armijo conditions:

	.. math::
		
		\mathbf{x}_{k+1} \leftarrow \mathbf{x}_k + \alpha \mathbf{p}_k.
		
	This implementation offers several features that modify this basic outline.

	First, the user can specify a nonlinear *trajectory* along which candidate points 
	can move; i.e.,

	.. math::

		\mathbf{x}_{k+1} \leftarrow T(\mathbf{x}_k, \mathbf{p}_k, \alpha). 
	
	Second, the user can specify a custom solver for computing the search direction :math:`\mathbf{p}_k`.

	Parameters
	----------
	f : callable
		residual, :math:`\mathbf{f}: \mathbb{R}^m \to \mathbb{R}^M`
	F : callable
		Jacobian of residual :math:`\mathbf{f}`; :math:`\mathbf{F}: \mathbb{R}^m \to \mathbb{R}^{M \times m}`
	tol: float [optional] default = 1e-8
		gradient norm stopping criterion
	tol_normdx: float [optional] default = 1e-12
		norm of control update stopping criterion
	maxiter : int [optional] default = 100
		maximum number of iterations of Gauss-Newton
	linesearch: callable, returns new x
		f : callable, residual, f: R^n -> R^m
		g : gradient, R^n
		p : descent direction, R^n
		x0 : current iterate, R^n
	gnsolver: [optional] callable, returns search direction p 
		Parameters: 
			F: current Jacobian
			f: current residual

		Returns:
			p: search step
			s: singular values of Jacobian
	verbose: int [optional] default = 0
		if >= print convergence history diagnostics

	Returns
	-------
	numpy.array((dof,))
		returns x^* (optimizer)
	int
		info = 0: converged with norm of gradient below tol
		info = 1: norm of gradient did not converge, but ||dx|| below tolerance
		info = 2: did not converge, max iterations exceeded
	"""
	n = len(x0)
	if maxiter <= 0: return x0, 4

	if verbose >= 1:
		print('Gauss-Newton Solver Iteration History')
		print('  iter   |   ||f(x)||   |   ||dx||   | cond(F(x)) |   alpha    |  ||grad||  ')
		print('---------|--------------|------------|------------|------------|------------')
	if trajectory is None:
		trajectory = lambda x0, p, t: x0 + t * p

	if linesearch is None:
		linesearch = linesearch_armijo
			
	if gnsolver is None:
		# Scipy seems to properly check for proper allocation of working space, reporting an error with gelsd
		# so we specify using gelss (an SVD based solver)
		def gnsolver(F_eval, f_eval):
			dx, _, _, s = sp.linalg.lstsq(F_eval, -f_eval, lapack_driver = 'gelss')
			return dx, s

	x = np.copy(x0)
	f_eval = f(x)
	F_eval = F(x)
	grad = F_eval.T @ f_eval

	normgrad = np.linalg.norm(grad)

	#rescale tol by norm of initial gradient
	tol = max(tol*normgrad, 1e-14)

	normdx = 1
	for it in range(maxiter):
		residual_increased = False
		
		# Compute search direction
		dx, s = gnsolver(F_eval, f_eval)
		
		# Check we got a valid search direction
		if not np.all(np.isfinite(dx)):
			raise RuntimeError("Non-finite search direction returned") 
		
		# If Gauss-Newton step is not a descent direction, use -gradient instead
		if np.inner(grad, dx) >= 0:
			dx = -grad
		
		# Back tracking line search
		x_new, alpha, f_eval_new = linesearch(f, grad, dx, x, trajectory=trajectory)
		

		normf = np.linalg.norm(f_eval_new)	
		if np.linalg.norm(f_eval_new) >= np.linalg.norm(f_eval):
			residual_increased = True
		else:
			#f_eval = f(x)
			f_eval = f_eval_new
			x = x_new

		normdx = np.linalg.norm(dx)
		F_eval = F(x)
		grad = F_eval.T @ f_eval_new
		

		#########################################################################
		# Printing section 
		if s[-1] == 0:
			cond = np.inf
		else:	
			cond = s[0] / s[-1]
		
		if verbose >= 1:
			normgrad = np.linalg.norm(grad)
			print(
				'    %3d  |  %1.4e  |  %8.2e  |  %8.2e  |  %8.2e  |  %8.2e' % (
				it, normf, normdx, cond, alpha, normgrad))
		# Termination conditions
		if normgrad < tol:
			if verbose: print("norm gradient %1.3e less than tolerance %1.3e" % 
				(normgrad, tol))
			break
		if normdx < tol_normdx:
			if verbose: print("norm dx %1.3e less than tolerance %1.3e" % 
				(normdx, tol_normdx))
			break
		if residual_increased:
			if verbose: print("residual increased during line search from %1.5e to %1.5e" % 
				(np.linalg.norm(f_eval), np.linalg.norm(f_eval_new)))
			break

	if normgrad <= tol:
		info = 0
		if verbose >= 1:
			print('Gauss-Newton converged successfully!')
	elif normdx <= tol_normdx:
		info = 1
		if verbose >= 1:
			print ('Gauss-Newton did not converge: ||dx|| < tol')
	elif it == maxiter - 1:
		info = 2
		if verbose >= 1:
			print ('Gauss-Newton did not converge: max iterations reached')
	elif np.linalg.norm(f_eval_new) >= np.linalg.norm(f_eval):
		info = 3
		if verbose >= 1:
			print ('No progress made during line search')
	else:
		raise Exception('Stopping criteria not determined!')

	return x, info





__all__ = ["make_sir"]

import grama as gr

from scipy.integrate import solve_ivp
from scipy.special import lambertw
from numpy import real
from toolz import curry

DF = gr.Intention()

## Model setup
def sir_rhs(t, y, beta, gamma):
    r"""Right-hand side (RHS) of SIR ODE model

    Implements the evolution equations (right-hand side of ODE system) for the SIR compartmental model for disease transmission.

    Args:
        t (float): Time (unused)
        y (array-like): State variables, [S, I, R]
        beta (float): Infectivity parameter
        gamma (float): Recovery parameter

    Returns:
        array-like: Time-derivatives of state variables, [dSdt, dIdt, dRdt]

    """
    ## Unpack data
    N = sum(y) # Total population
    S, I, R = y

    ## Compute derivatives
    dSdt = -beta * I * S / N
    dIdt = +beta * I * S / N - gamma * I
    dRdt = +gamma * I

    ## Package for solver
    return [dSdt, dIdt, dRdt]

def sir_vtime(T, S0, I0, R0, beta, gamma, rtol=1e-4):
    r"""Solve SIR IVP, vectorized over T

    Solves the initial value problem (IVP) associated with the SIR model, given parameter values and a span of time values. This routine uses an adaptive timestep to solve the IVP to a specified tolerance, then uses cubic interpolation to query the time points of interest.

    Args:
        T (array-like): Time points of interest
        S0 (float): Initial number of susceptible individuals (at t=0)
        I0 (float): Initial number of infected individuals (at t=0)
        R0 (float): Initial number of removed individuals (at t=0)
        beta (float): Infection rate parameter
        gamma (float): Removal rate parameter

    Returns:
        pandas DataFrame: Simulation timeseries results
    """

    ## Solve SIR model on adaptive, coarse time mesh
    T_span = [0, max(T)]
    y0 = [S0, I0, R0]

    res = solve_ivp(
        sir_rhs,
        T_span,
        y0,
        args=(beta, gamma),
        rtol=rtol,
        t_eval=T,
    )

    ## Interpolate to desired T points
    df_res = gr.df_make(
        t=T,
        S=res.y[0, :],
        I=res.y[1, :],
        R=res.y[2, :],
        S0=[S0],
        I0=[I0],
        R0=[R0],
        beta=[beta],
        gamma=[gamma],
    )

    return df_res

@curry
def fun_sir(df, rtol=1e-4):
    r"""Fully-vectorized SIR solver

    SIR IVP solver, vectorized over parameter values **and** time. The routine identifies groups of parameter values and runs a vectorized IVP solver over all associated time points, and gathers all results into a single output DataFrame. Intended for use in a grama model.

    Args:
        df (pd.DataFrame): All input values; must contain columns for all SIR parameters

    Preconditions:
        ["t", "S0", "I0", "R0", "beta", "gamma"] in df.columns

    Postconditions:
        Row-ordering of input data is reflected in the row-ordering of the output.

    Returns:
        pd.DataFrame: Solution results
    """
    ## Find all groups of non-t parameters
    df_grouped = (
        df
        >> gr.tf_mutate(
            _idx=DF.index,
            _code=gr.str_c(
                "S0", DF.S0,
                "I0", DF.I0,
                "R0", DF.R0,
                "beta", DF.beta,
                "gamma", DF.gamma,
            )
        )
    )

    ## Run time-vectorized SIR solver over each group
    df_results = gr.df_grid()
    codes = set(df_grouped._code)
    for code in codes:
        df_param = (
            df_grouped
            >> gr.tf_filter(DF._code == code)
        )

        df_results = (
            df_results
            >> gr.tf_bind_rows(
                sir_vtime(
                    df_param.t,
                    df_param.S0[0],
                    df_param.I0[0],
                    df_param.R0[0],
                    df_param.beta[0],
                    df_param.gamma[0],
                    rtol=rtol,
                )
                >> gr.tf_mutate(_idx=df_param._idx)
            )
        )

    ## Sort to match original ordering
    # NOTE: Without this, the output rows will be scrambled, relative
    # to the input rows, leading to very confusing output!
    return (
        df_results
        >> gr.tf_arrange(DF._idx)
        >> gr.tf_drop("_idx")
    )

## Model builder
def make_sir(rtol=1e-4):
    r"""Make an SIR model

    Instantiates a Susceptible, Infected, Removed (SIR) model for disease transmission.

    Args:
        rtol (float): Relative tolerance for IVP solver

    Returns:
        grama Model: SIR model

    References:
        "Compartmental models in epidemiology," Wikipedia, url: https://en.wikipedia.org/wiki/Compartmental_models_in_epidemiology

    """

    md_sir = (
        gr.Model("SIR Model")
        >> gr.cp_vec_function(
            fun=lambda df: gr.df_make(
                # Assume no recovered people at virus onset
                R0=0,
                # Assume total population of N=100
                S0=df.N - df.I0,
            ),
            var=["I0", "N"],
            out=["S0", "R0"],
            name="Population setup",
        )
        >> gr.cp_vec_function(
            fun=fun_sir(rtol=rtol),
            var=["t", "S0", "I0", "R0", "beta", "gamma"],
            out=["S", "I", "R"],
            name="ODE solver & interpolation",
        )
        >> gr.cp_bounds(
            N=(100, 100), # Fixed population size
            I0=(1, 10),
            beta=(0.1, 0.5),
            gamma=(0.1, 0.5),
            t=(0, 100),
        )
    )

    return md_sir



__all__ = [
    "eval_sinews",
    "ev_sinews",
    "eval_hybrid",
    "ev_hybrid",
]

from grama import add_pipe, CopulaIndependence, custom_formatwarning
from grama import comp_marginals, eval_df
from grama.eval_defaults import invariants_eval_model, invariants_eval_df
from numbers import Integral
from numpy import tile, linspace, zeros, isfinite
from numpy.linalg import cholesky, inv
from numpy.random import random
from numpy.random import seed as set_seed
from pandas import DataFrame
from scipy.stats import norm, lognorm
from toolz import curry
from warnings import formatwarning, catch_warnings, simplefilter

formatwarning = custom_formatwarning


## Marginal sweeps with random origins
# --------------------------------------------------
@curry
def eval_sinews(
    model,
    n_density=10,
    n_sweeps=3,
    seed=None,
    df_det=None,
    varname="sweep_var",
    indname="sweep_ind",
    append=True,
    skip=False,
):
    r"""Sweep study

    Perform coordinate sweeps over each model random variable ("sinew" design). Use random starting points drawn from the joint density. Optionally sweep the deterministic variables.

    For more expensive models, it can be helpful to tune n_density and n_sweeps to achieve a reasonable runtime.

    Use gr.plot_auto() to construct a quick visualization of the output dataframe. Use `skip` version to visualize the design, and non-skipped version to visualize the results.

    Args:
        model (gr.Model): Model to evaluate
        n_density (numeric): Number of points along each sweep
        n_sweeps (numeric): Number of sweeps per-random variable
        seed (int): Random seed to use
        df_det (DataFrame): Deterministic levels for evaluation;
            use "nom" for nominal deterministic levels,
            use "swp" to sweep deterministic variables
        varname (str): Column name to give for sweep variable; default="sweep_var"
        indname (str): Column name to give for sweep index; default="sweep_ind"
        append (bool): Append results to conservative inputs?
        skip (bool): Skip evaluation of the functions?

    Returns:
        DataFrame: Results of evaluation or unevaluated design

    Examples::

        import grama as gr
        md = gr.make_cantilever_beam()
        # Skip evaluation, used to visualize the design (input points)
        df_design = md >> gr.ev_sinews(df_det="nom", skip=True)
        df_design >> gr.pt_auto()
        # Visualize the input-to-output relationships of the model
        df_sinew = md >> gr.ev_sinews(df_det="nom")
        df_sinew >> gr.pt_auto()

    """
    ## Common invariant checks
    invariants_eval_model(model, skip)
    invariants_eval_df(df_det, arg_name="df_det", valid_str=["nom", "swp"])
    ## Override model if deterministic sweeps desired
    if df_det == "swp":
        ## Collect sweep-able deterministic variables
        var_sweep = list(
            filter(
                lambda v: isfinite(model.domain.get_width(v))
                & (model.domain.get_width(v) > 0),
                model.var_det,
            )
        )
        ## Generate pseudo-marginals
        dicts_var = {}
        for v in var_sweep:
            dicts_var[v] = {
                "dist": "uniform",
                "loc": model.domain.get_bound(v)[0],
                "scale": model.domain.get_width(v),
            }
        ## Overwrite model
        model = comp_marginals(model, **dicts_var)
        ## Restore flag
        df_det = "nom"

    ## Set seed only if given
    if seed is not None:
        set_seed(seed)

    ## Ensure sample count is int
    if not isinstance(n_density, Integral):
        print("eval_sinews() is rounding n_density...")
        n_density = int(n_density)
    if not isinstance(n_sweeps, Integral):
        print("eval_sinews() is rounding n_sweeps...")
        n_sweeps = int(n_sweeps)

    ## Build quantile sweep data
    q_random = tile(random((1, model.n_var_rand, n_sweeps)), (n_density, 1, 1))
    q_dense = linspace(0, 1, num=n_density)
    Q_all = zeros((n_density * n_sweeps * model.n_var_rand, model.n_var_rand))
    C_var = ["tmp"] * (n_density * n_sweeps * model.n_var_rand)
    C_ind = [0] * (n_density * n_sweeps * model.n_var_rand)

    ## Interlace
    for i_input in range(model.n_var_rand):
        ind_base = i_input * n_density * n_sweeps
        for i_sweep in range(n_sweeps):
            ind_start = ind_base + i_sweep * n_density
            ind_end = ind_base + (i_sweep + 1) * n_density

            Q_all[ind_start:ind_end] = q_random[:, :, i_sweep]
            Q_all[ind_start:ind_end, i_input] = q_dense
            C_var[ind_start:ind_end] = [model.var_rand[i_input]] * n_density
            C_ind[ind_start:ind_end] = [i_sweep] * n_density

            ## Modify endpoints for infinite support
            if not isfinite(model.density.marginals[model.var_rand[i_input]].q(0)):
                Q_all[ind_start, i_input] = 1 / n_density / 10
            if not isfinite(model.density.marginals[model.var_rand[i_input]].q(1)):
                Q_all[ind_end - 1, i_input] = 1 - 1 / n_density / 10

    ## Assemble sampling plan
    df_pr = DataFrame(data=Q_all, columns=model.var_rand)
    df_rand = model.density.pr2sample(df_pr)
    df_rand[varname] = C_var
    df_rand[indname] = C_ind
    ## Construct outer-product DOE
    df_samp = model.var_outer(df_rand, df_det=df_det)

    if skip:
        ## Evaluation estimate
        runtime_msg = model.runtime_message(df_samp)
        print(runtime_msg)

        ## For autoplot
        with catch_warnings():
            simplefilter("ignore")
            df_samp._plot_info = {"type": "sinew_inputs", "var": model.var_rand}

        ## Pass-through
        return df_samp

    ## Apply
    df_res = eval_df(model, df=df_samp, append=append)
    ## For autoplot
    with catch_warnings():
        simplefilter("ignore")
        df_res._plot_info = {
            "type": "sinew_outputs",
            "var": model.var_rand,
            "out": model.out,
        }

    return df_res


ev_sinews = add_pipe(eval_sinews)

## Hybrid points for Sobol' indices
# --------------------------------------------------
@curry
def eval_hybrid(
    model,
    n=1,
    plan="first",
    df_det=None,
    varname="hybrid_var",
    seed=None,
    append=True,
    skip=False,
):
    r"""Hybrid points for Sobol' indices

    Use the "hybrid point" design (Sobol', 1999) to support estimating Sobol'
    indices. Use gr.tran_sobol() to post-process the results and compute
    estimates.

    Args:
        model (gr.Model): Model to evaluate; must have CopulaIndependence
        n (numeric): Number of points along each sweep
        plan (str): Sobol' index to compute; plan={"first", "total"}
        seed (int): Random seed to use
        df_det (DataFrame): Deterministic levels for evaluation; use "nom"
            for nominal deterministic levels.
        varname (str): Column name to give for sweep variable; default="hybrid_var"
        append (bool): Append results to conservative inputs?
        skip (bool): Skip evaluation of the functions?

    Returns:
        DataFrame: Results of evaluation or unevaluated design

    References:
        I.M. Sobol', "Sensitivity Estimates for Nonlinear Mathematical Models" (1999) MMCE, Vol 1.

    Examples::

        import grama as gr
        md = gr.make_cantilever_beam()
        ## Compute the first-order indices
        df_first = md >> gr.ev_hybrid(df_det="nom", plan="first")
        df_first >> gr.tf_sobol()
        ## Compute the total-order indices
        df_total = md >> gr.ev_hybrid(df_det="nom", plan="total")
        df_total >> gr.tf_sobol()

    """
    ## Check invariants
    invariants_eval_model(model, skip)
    invariants_eval_df(df_det, arg_name="df_det", valid_str=["nom"])
    if not isinstance(model.density.copula, CopulaIndependence):
        raise ValueError(
            "model must have CopulaIndependence structure;\n"
            + "Sobol' indices only defined for independent variables"
        )

    ## Set seed only if given
    if seed is not None:
        set_seed(seed)

    if not isinstance(n, Integral):
        print("eval_hybrid() is rounding n...")
        n = int(n)

    ## Draw hybrid points
    X = random((n, model.n_var_rand))
    Z = random((n, model.n_var_rand))

    ## Reserve space
    Q_all = zeros((n * (model.n_var_rand + 1), model.n_var_rand))
    Q_all[:n] = X  # Base samples
    C_var = ["_"] * (n * (model.n_var_rand + 1))

    ## Interleave samples
    for i_in in range(model.n_var_rand):
        i_start = (i_in + 1) * n
        i_end = (i_in + 2) * n

        if plan == "first":
            Q_all[i_start:i_end, :] = Z
            Q_all[i_start:i_end, i_in] = X[:, i_in]
        elif plan == "total":
            Q_all[i_start:i_end, :] = X
            Q_all[i_start:i_end, i_in] = Z[:, i_in]
        else:
            raise ValueError("plan must be `first` or `total`")

        C_var[i_start:i_end] = [model.var_rand[i_in]] * n

    ## Construct sampling plan
    df_pr = DataFrame(data=Q_all, columns=model.var_rand)
    ## Convert samples to desired marginals
    df_rand = model.density.pr2sample(df_pr)
    df_rand[varname] = C_var
    ## Construct outer-product DOE
    df_samp = model.var_outer(df_rand, df_det=df_det)

    if skip:
        ## Evaluation estimate
        runtime_msg = model.runtime_message(df_samp)
        print(runtime_msg)

        # Generate design
        with catch_warnings():
            simplefilter("ignore")
            df_samp._meta = dict(
                type="eval_hybrid",
                varname=varname,
                plan=plan,
                var_rand=model.var_rand,
                out=model.out,
            )

        return df_samp

    df_res = eval_df(model, df=df_samp, append=append)
    with catch_warnings():
        simplefilter("ignore")
        df_res._meta = dict(
            type="eval_hybrid",
            varname=varname,
            plan=plan,
            var_rand=model.var_rand,
            out=model.out,
        )

    return df_res


ev_hybrid = add_pipe(eval_hybrid)



__all__ = [
    "tran_arrange",
    "tf_arrange",
    "tran_rename",
    "tf_rename",
    "tran_separate",
    "tf_separate",
    "tran_unite",
    "tf_unite",
    "tran_gather",
    "tf_gather",
    "tran_spread",
    "tf_spread",
    "tran_explode",
    "tf_explode",
    "convert_type"
]

import re
from .base import dfdelegate, symbolic_evaluation, flatten
from .. import add_pipe
from numpy import arange, nan
from pandas import Series, concat, melt, to_numeric, to_datetime


# ------------------------------------------------------------------------------
# Sorting
# ------------------------------------------------------------------------------


@dfdelegate
def tran_arrange(df, *args, **kwargs):
    """Calls `pandas.DataFrame.sort_values` to sort a DataFrame according to
    criteria.

    See:
    http://pandas.pydata.org/pandas-docs/stable/generated/pandas.DataFrame.sort_values.html

    For a list of specific keyword arguments for sort_values (which will be
    the same in arrange).

    Args:
        *args: Symbolic, string, integer or lists of those types indicating
            columns to sort the DataFrame by.

    Kwargs:
        **kwargs: Any keyword arguments will be passed through to the pandas
            `DataFrame.sort_values` function.
    """

    flat_args = [a for a in flatten(args)]

    series = [
        df[arg]
        if isinstance(arg, str)
        else df.iloc[:, arg]
        if isinstance(arg, int)
        else Series(arg)
        for arg in flat_args
    ]

    sorter = concat(series, axis=1).reset_index(drop=True)
    sorter = sorter.sort_values(sorter.columns.tolist(), **kwargs)
    return df.iloc[sorter.index, :].reset_index(drop=True)

tf_arrange = add_pipe(tran_arrange)


# ------------------------------------------------------------------------------
# Renaming
# ------------------------------------------------------------------------------


@dfdelegate
@symbolic_evaluation(eval_as_label=True)
def tran_rename(df, **kwargs):
    """Renames columns

    Renames columns, where keyword argument values are the current names of columns and keys are the new names.

    You can think of the keyword argument values as `newname="oldname"`; note that new variable names must follow Python variable naming conventions (no spaces, names can't start with numbers, etc.). See the Examples section below for an example of the renaming syntax.

    Args:
        df (pandas.DataFrame): DataFrame

    Kwargs:
        **kwargs: Renaming pair
            the name of the argument (left of `=`) will be the new column name,
            the value of the argument (right of `=`) is the old column name (as a string).

    Examples::

        ## Setup
        import grama as gr
        DF = gr.Intention()
        ## Load example dataset
        from grama.data import df_stang

        ## Rename columns
        (
            df_stang
            >> gr.tf_rename(
                ## Remember, the pattern is new="old"
                thickness="thick",
                alloy_name="alloy",
                elasticity="E",
                poissons_ratio="mu",
                measurement_angle="ang",
            )
        )

    """

    return df.rename(columns={v: k for k, v in kwargs.items()})

tf_rename = add_pipe(tran_rename)


# ------------------------------------------------------------------------------
# Elongate
# ------------------------------------------------------------------------------


@dfdelegate
@symbolic_evaluation(eval_as_label=["*"])
def tran_gather(df, key, values, *args, **kwargs):
    """
    Melts the specified columns in your DataFrame into two key:value columns.

    Args:
        key (str): Name of identifier column.
        values (str): Name of column that will contain values for the key.
        *args (str, int, symbolic): Columns to "melt" into the new key and
            value columns. If no args are specified, all columns are melted
            into they key and value columns.

    Kwargs:
        add_id (bool): Boolean value indicating whether to add a `"_ID"`
            column that will preserve information about the original rows
            (useful for being able to re-widen the data later).

    Example ::

        diamonds >> gather('variable', 'value', ['price', 'depth','x','y','z']) >> head(5)

           carat      cut color clarity  table variable  value
        0   0.23    Ideal     E     SI2   55.0    price  326.0
        1   0.21  Premium     E     SI1   61.0    price  326.0
        2   0.23     Good     E     VS1   65.0    price  327.0
        3   0.29  Premium     I     VS2   58.0    price  334.0
        4   0.31     Good     J     SI2   58.0    price  335.0
    """

    if len(args) == 0:
        args = df.columns.tolist()
    else:
        args = [a for a in flatten(args)]

    if kwargs.get("add_id", False):
        df = df.assign(_ID=arange(df.shape[0]))

    columns = df.columns.tolist()
    id_vars = [col for col in columns if col not in args]
    return melt(df, id_vars, list(args), key, values)

tf_gather = add_pipe(tran_gather)


# ------------------------------------------------------------------------------
# Widen
# ------------------------------------------------------------------------------


def convert_type(df, columns):
    """
    Helper function that attempts to convert columns into their appropriate
    data type.
    """
    # taken in part from the dplython package
    out_df = df.copy()
    for col in columns:
        column_values = Series(out_df[col].unique())
        column_values = column_values[~column_values.isnull()]
        # empty
        if len(column_values) == 0:
            continue
        # boolean
        if set(column_values.values) < {"True", "False"}:
            out_df[col] = out_df[col].map({"True": True, "False": False})
            continue
        # numeric
        if to_numeric(column_values, errors="coerce").isnull().sum() == 0:
            out_df[col] = to_numeric(out_df[col], errors="ignore")
            continue
        # datetime
        if to_datetime(column_values, errors="coerce").isnull().sum() == 0:
            out_df[col] = to_datetime(
                out_df[col], errors="ignore", infer_datetime_format=True
            )
            continue

    return out_df


@dfdelegate
@symbolic_evaluation(eval_as_label=["*"])
def tran_spread(df, key, values, convert=False, fill=None):
    """
    Transforms a "long" DataFrame into a "wide" format using a key and value
    column.

    If you have a mixed datatype column in your long-format DataFrame then the
    default behavior is for the spread columns to be of type `object`, or
    string. If you want to try to convert dtypes when spreading, you can set
    the convert keyword argument in spread to True.

    Args:
        key (str, int, or symbolic): Label for the key column.
        values (str, int, or symbolic): Label for the values column.

    Kwargs:
        convert (bool): Boolean indicating whether or not to try and convert
            the spread columns to more appropriate data types.


    Examples::

        widened = elongated >> spread(X.variable, X.value)
        widened >> head(5)

            _ID carat clarity color        cut depth price table     x     y     z
        0     0  0.23     SI2     E      Ideal  61.5   326    55  3.95  3.98  2.43
        1     1  0.21     SI1     E    Premium  59.8   326    61  3.89  3.84  2.31
        2    10   0.3     SI1     J       Good    64   339    55  4.25  4.28  2.73
        3   100  0.75     SI1     D  Very Good  63.2  2760    56   5.8  5.75  3.65
        4  1000  0.75     SI1     D      Ideal  62.3  2898    55  5.83   5.8  3.62
    """

    # Taken mostly from dplython package
    columns = df.columns.tolist()
    id_cols = [col for col in columns if not col in [key, values]]

    temp_index = ["" for i in range(len(df))]
    for id_col in id_cols:
        temp_index += df[id_col].map(str)

    out_df = df.assign(temp_index=temp_index)
    out_df = out_df.set_index("temp_index")
    spread_data = out_df[[key, values]]

    if not all(
        spread_data.groupby([spread_data.index, key]).agg("count").reset_index()[values]
        < 2
    ):
        raise ValueError("Duplicate identifiers")

    spread_data = spread_data.pivot(columns=key, values=values)

    if convert and (out_df[values].dtype.kind in "OSaU"):
        columns_to_convert = [col for col in spread_data if col not in columns]
        spread_data = convert_type(spread_data, columns_to_convert)

    if not (fill is None):
        spread_data.fillna(value=fill, inplace=True)

    out_df = out_df[id_cols].drop_duplicates()
    out_df = out_df.merge(spread_data, left_index=True, right_index=True).reset_index(
        drop=True
    )

    out_df = (out_df >> tf_arrange(id_cols)).reset_index(drop=True)

    return out_df

tf_spread = add_pipe(tran_spread)


# ------------------------------------------------------------------------------
# Separate columns
# ------------------------------------------------------------------------------


@dfdelegate
@symbolic_evaluation(eval_as_label=["*"])
def tran_separate(
    df,
    column,
    into,
    sep="[\W_]+",
    remove=True,
    convert=False,
    extra="drop",
    fill="right",
):
    """Splits one column into multiple columns.

    Split a single column containing string values into multiple columns by *separating* the strings on a specified pattern. The separator pattern is specified via the `sep` argument, and may be a regular expression.

    This verb is often used in conjunction with pivoting (e.g. tran_pivot_longer()) to separate column names from data.

    Args:
        df (pandas.DataFrame): DataFrame passed in through the pipe.
        column (str, symbolic): Label of column to split.
        into (list): List of string names for new columns.

    Kwargs:
        sep (str or list): If a string, the regex string used to split the
            column. If a list, a list of integer positions to split strings
            on.
        remove (bool): Boolean indicating whether to remove the original column.
        convert (bool): Boolean indicating whether the new columns should be converted to the appropriate type.
        extra (str): either `'drop'`, where split pieces beyond the specified new columns are dropped, or `'merge'`, where the final split piece contains the remainder of the original column.
        fill (str): either `'right'`, where `np.nan` values are filled in the right-most columns for missing pieces, or `'left'` where `np.nan` values are filled in the left-most columns.

    Returns:
        pandas.DataFrame: Modified data

    Examples::

        import grama as gr
        DF = gr.Intention

        ## Simple example
        df = gr.df_make(x=["a_1", "b_2", "c_3"])
        (
            df
            >> gr.tf_separate(
                column=DF.x,
                into=["letter", "number"],
                sep="_",
            )
        )

        ## tran_separate is helpful when pivoting data
        from grama.data import df_stang_wide
        (
            df_stang_wide
            >> gr.tf_pivot_longer(
                columns=["E_00", "mu_00", "E_45", "mu_45", "E_90", "mu_90"],
                names_to="name",
                values_to="value",
            )
            >> gr.tf_separate(
                column=DF.name,
                into=["variable", "angle"],
                sep="_",
            )
        (
    """

    assert isinstance(into, (tuple, list))

    if isinstance(sep, (tuple, list)):
        inds = [0] + list(sep)
        if len(inds) > len(into):
            if extra == "drop":
                inds = inds[: len(into) + 1]
            elif extra == "merge":
                inds = inds[: len(into)] + [None]
        else:
            inds = inds + [None]

        splits = df[column].map(
            lambda x: [
                str(x)[slice(inds[i], inds[i + 1])] if i < len(inds) - 1 else nan
                for i in range(len(into))
            ]
        )

    else:
        maxsplit = len(into) - 1 if extra == "merge" else 0
        splits = df[column].map(lambda x: re.split(sep, x, maxsplit))

    right_filler = lambda x: x + [nan for i in range(len(into) - len(x))]
    left_filler = lambda x: [nan for i in range(len(into) - len(x))] + x

    if fill == "right":
        splits = [right_filler(x) for x in splits]
    elif fill == "left":
        splits = [left_filler(x) for x in splits]

    for i, split_col in enumerate(into):
        df[split_col] = [x[i] if not x[i] == "" else nan for x in splits]

    if convert:
        df = convert_type(df, into)

    if remove:
        df.drop(column, axis=1, inplace=True)

    return df

tf_separate = add_pipe(tran_separate)


# ------------------------------------------------------------------------------
# Unite columns
# ------------------------------------------------------------------------------


@dfdelegate
@symbolic_evaluation(eval_as_label=["*"])
def tran_unite(df, colname, *args, **kwargs):
    """Unite multiple columns into one

    Does the inverse of `tran_separate`; joins columns together by a specified separator.

    Any columns that are not strings will be converted to strings.

    Args:
        df (pandas.DataFrame): DataFrame passed in through the pipe.
        colname (str): the name of the new joined column.
        *args: list of columns to be joined, which can be strings, symbolic, or
            integer positions.

    Kwargs:
        sep (str): the string separator to join the columns with.
        remove (bool): Boolean indicating whether or not to remove the
            original columns.
        na_action (str): can be one of `'maintain'` (the default),
            '`ignore'`, or `'as_string'`. The default will make the new column
            row a `NaN` value if any of the original column cells at that
            row contained `NaN`. '`ignore'` will treat any `NaN` value as an
            empty string during joining. `'as_string'` will convert any `NaN`
            value to the string `'nan'` prior to joining.

    """

    to_unite = list([a for a in flatten(args)])
    sep = kwargs.get("sep", "_")
    remove = kwargs.get("remove", True)
    # possible na_action values
    # ignore: empty string
    # maintain: keep as np.nan (default)
    # as_string: becomes string 'nan'
    na_action = kwargs.get("na_action", "maintain")

    # print(to_unite, sep, remove, na_action)

    if na_action == "maintain":
        df[colname] = df[to_unite].apply(
            lambda x: nan if any(x.isnull()) else sep.join(x.map(str)), axis=1
        )
    elif na_action == "ignore":
        df[colname] = df[to_unite].apply(
            lambda x: sep.join(x[~x.isnull()].map(str)), axis=1
        )
    elif na_action == "as_string":
        df[colname] = df[to_unite].astype(str).apply(lambda x: sep.join(x), axis=1)

    if remove:
        df.drop(to_unite, axis=1, inplace=True)

    return df

tf_unite = add_pipe(tran_unite)


# ------------------------------------------------------------------------------
# Nesting
# ------------------------------------------------------------------------------


@dfdelegate
@symbolic_evaluation(eval_as_label=["*"])
def tran_explode(df, col, convert=False):
    """Lengthen DataFrame by exploding iterable entries in a column.

    If you have a mixed datatype column in your long-format DataFrame then the
    default behavior is for the spread columns to be of type `object`, or
    string. If you want to try to convert dtypes when spreading, you can set
    the convert keyword argument in spread to True.

    Note: Implemented in terms of the Pandas pd.DataFrame.explode() function.

    Args:
        col (str, int, or symbolic): Label for the column to explode.

    Kwargs:
        convert (bool): Boolean indicating whether or not to try and convert
            the spread columns to more appropriate data types.

    Returns:
        DataFrame:

    Example::

    """

    df_res = df.explode(col).reset_index(drop=True)

    if convert and (df_res[col].dtype.kind in "OSaU"):
        return convert_type(df_res, [col])
    return df_res

tf_explode = add_pipe(tran_explode)



""" Utility containing various geometric routines, most of which are used in sampling
"""
from __future__ import print_function
import numpy as np
from scipy.spatial import Voronoi 
from scipy.spatial.distance import cdist, pdist, squareform

__all__ = ['sample_sphere', 'unique_points', 'sample_simplex']




def sample_sphere(dim, n, k = 100):
	""" Sample points on a high-dimensional sphere 

	Uses Mitchell's best candidate algorithm to obtain a 
	quasi-uniform distribution of points on the sphere,


	See:
		https://www.jasondavies.com/maps/random-points/
		https://bl.ocks.org/mbostock/d7bf3bd67d00ed79695b

	Parameters
	----------
	dim: int, positive
		Dimension of the space to sample
	n: int, positive
		Number of points to sample
	k: int, positive (optional)
		Number of candidates to take at each step
	"""
	X = np.zeros( (n, dim) )
	
	# First sample
	x = np.random.randn(dim)
	x /= np.linalg.norm(x)
	X[0] = x
	
	for i in range(1,n):
		# Draw candidates (normalized points on the sphere)
		Xcan = np.random.randn(k, dim)
		Xcan = (Xcan.T/np.sqrt(np.sum(Xcan**2, axis = 1))).T

		# Compute the distance
		dist = np.min(1 - np.dot(X[:i,], Xcan.T), axis = 0)
		I = np.argmax(dist)
		X[i] = Xcan[I]

	return X

def unique_points(X):
	r""" Compute the unique points from a list

	Parameters
	----------
	X: array-like (M, m)
		Input points
	
	Returns
	-------
	I: np.ndarray (M, dtype = np.bool)
		List of points with no points close to each other
	"""
	D = squareform(pdist(X))
	I = np.ones(len(X), dtype = np.bool)
	for i in range(len(X)-1):
		I[i] = ~np.isclose(np.min(D[i,i+1:]), 0)
	return I


def sample_simplex(dim, Nsamp = 1):
	r""" Samples the unit simplex uniformly



	References:
	https://cs.stackexchange.com/questions/3227/uniform-sampling-from-a-simplex

	"""
	r = np.random.uniform(0, 1, size = (Nsamp, dim+1))
	r[:,0] = 0
	r[:,-1] = 1
	r = np.sort(r, axis = 1)
	alphas = r[:,1:] - r[:,:-1]
	return alphas



import unittest

from context import grama as gr
from context import data
DF = gr.Intention()

## Core function tests
##################################################
class TestSPC(unittest.TestCase):
    """Test statistical process control (SPC) tools
    """

    def test_factors(self):
        """Test factor implementations"""
        ## Standard deviation de-biasing constant
        self.assertAlmostEqual(
            gr.c_sd(2),
            0.7979,
            places=4,
        )
        self.assertAlmostEqual(
            gr.c_sd(4),
            0.9213,
            places=4,
        )
        self.assertAlmostEqual(
            gr.c_sd(10),
            0.9727,
            places=4,
        )
        ## Standard deviation control limits
        self.assertAlmostEqual(
            gr.B3(5),
            0.0000,
            places=3,
        )
        self.assertAlmostEqual(
            gr.B3(6),
            0.030,
            places=3,
        )
        self.assertAlmostEqual(
            gr.B3(10),
            0.284,
            places=3,
        )

        self.assertAlmostEqual(
            gr.B4(5),
            2.089,
            places=3,
        )
        self.assertAlmostEqual(
            gr.B4(6),
            1.970,
            places=3,
        )
        self.assertAlmostEqual(
            gr.B4(10),
            1.716,
            places=3,
        )

    def test_plot_xbs(self):
        r"""Tests that Xbar and S chart runs"""
        ## Basic functionality
        (
            data.df_shewhart
            >> gr.tf_mutate(idx=DF.index // 10)
            >> gr.pt_xbs(group="idx", var="tensile_strength")
        )

        ## Works with discrete group variable
        (
            data.df_shewhart
            >> gr.tf_mutate(idx=gr.as_factor(DF.index // 10))
            >> gr.pt_xbs(group="idx", var="tensile_strength")
        )

        ## Black & White color functionality
        (
            data.df_shewhart
            >> gr.tf_mutate(idx=DF.index // 10)
            >> gr.pt_xbs(group="idx", var="tensile_strength", color="bw")
        )



import numpy as np
import cvxpy as cp


def miniball(X, L = None):
	r""" Compute the smallest enclosing sphere

	TODO: Use implementation of https://github.com/hbf/miniball

	Parameters
	----------
	X: array-like (M,m)
		Points to enclose in a ball
	L: optional, (m,m)
		Lipschitz-like weighting metric
	
	Returns
	-------
	x: np.array(m)
		Center of circle
	r: float
		radius of circle
	"""
	X = np.array(X)
	M, m = X.shape
	if L is None:
		L = np.eye(m)

	x = cp.Variable(m)
	ones = np.ones((1, M))
	obj = cp.mixed_norm( (L @ ( cp.reshape(x,(m,1)) @ ones - X.T)).T, 2, np.inf)
	prob = cp.Problem(cp.Minimize(obj))
	prob.solve()

	return x.value, obj.value



__all__ = [
    "tran_head",
    "tf_head",
    "tran_tail",
    "tf_tail",
    "tran_sample",
    "tf_sample",
    "tran_distinct",
    "tf_distinct",
    "tran_row_slice",
    "tf_row_slice",
    "tran_filter",
    "tf_filter",
    "tran_top_n",
    "tf_top_n",
    "tran_pull",
    "tf_pull",
    "tran_dropna",
    "tf_dropna",
]

import warnings
from .base import dfdelegate, symbolic_evaluation, group_delegation
from .. import add_pipe
from numpy import array, ones
from pandas import Series


# ------------------------------------------------------------------------------
# `head` and `tail`
# ------------------------------------------------------------------------------


@dfdelegate
def tran_head(df, n=5):
    r"""Return the first n rows of a DataFrame
    """
    return df.head(n)

tf_head = add_pipe(tran_head)


@dfdelegate
def tran_tail(df, n=5):
    r"""Return the last n rows of a DataFrame
    """
    return df.tail(n)

tf_tail = add_pipe(tran_tail)


# ------------------------------------------------------------------------------
# Sampling
# ------------------------------------------------------------------------------


@dfdelegate
def tran_sample(df, *args, **kwargs):
    r"""Return a random subset of a DataFrame

    Arguments:
        n (int): Number of rows to return. Cannot be used with `frac`.
        frac (float): Fraction of items to return. Cannot be used with `n`.

    Returns:
        DataFrame: A random subset of rows from the original data.

    Notes:
        Alias for pandas DataFrame.sample(). See the docs for DataFrame.sample() for more information

    Examples:

        ## Setup
        import grama as gr
        ## Load example dataset
        from grama.data import df_diamonds
        ## Sample based on count
        (
            df_diamonds
            >> gr.tf_sample(n=100)
        )
        ## Sample based on fraction
        (
            df_diamonds
            >> gr.tf_sample(frac=0.01)
        )

    """
    return df.sample(*args, **kwargs).reset_index(drop=True)

tf_sample = add_pipe(tran_sample)


@group_delegation
@symbolic_evaluation(eval_as_label=["*"])
def tran_distinct(df, *args, **kwargs):
    r"""Return unique rows of a DataFrame

    Return unique rows of a DataFrame, determining uniqueness according to a user-defined set of columns.

    Args:
        *args (str or gr.Intention): Columns to consider when determining uniqueness. If empty, all columns are used for determining unique rows.

    Kwargs:
        keep ('first', 'last', False): Determines which, if any, duplicates to keep. 'first' returns the first occurrence, 'last' keeps the last occurrence, and `False` removes all duplicates.

    Returns:
        DataFrame: Data with unique rows

    Notes:
        Alias for pandas DataFrame.drop_duplicates(). See the docs for DataFrame.drop_duplicates for more information.

    Examples:
        ## Setup
        import grama as gr
        DF = gr.Intention()
        ## Load example dataset
        from grama.data import df_stang
        ## Find distinct thicknesses
        (
            df_stang
            >> gr.tf_distinct(DF.thick)
        )
        ## Find distinct combinations of thickness and angle
        (
            df_stang
            >> gr.tf_distinct(DF.thick, DF.angle)
        )
        ## Find all distinct rows
        (
            df_stang
            >> gr.tf_distinct()
        )
    """
    if not args:
        return df.drop_duplicates(**kwargs).reset_index(drop=True)
    return df.drop_duplicates(list(args), **kwargs).reset_index(drop=True)

tf_distinct = add_pipe(tran_distinct)


@dfdelegate
def tran_row_slice(df, indices):
    if isinstance(indices, (tuple, list)):
        indices = array(indices)
    if isinstance(indices, int):
        indices = array([indices])
    if isinstance(indices, Series):
        indices = indices.values

    if indices.dtype == bool:
        return df.loc[indices, :].reset_index(drop=True)
    return df.iloc[indices, :].reset_index(drop=True)

tf_row_slice = add_pipe(tran_row_slice)


# ------------------------------------------------------------------------------
# Filtering/masking
# ------------------------------------------------------------------------------


@dfdelegate
def mask(df, *args):
    r"""Select rows based on provided conditions

    Select for rows based on logical conditions. This can include equality between columns `DF.x == DF.y`, comparisons with a threshold `0 <= DF.z`, or other expressions that return a boolean value.

    Use the Intention operator (usually `DF = gr.Intention()`) as a convenient way to access columns in the DataFrame.

    There are a number of helper functions that make working with filters easier / more powerful. See also:

    - var_in() : Check if given value is one of a set of values
    - is_nan() : Check if given value is not a number (nan)
    - not_nan() : Check if given value is *not* not a number (nan)
    - str_detect() : Check for the presence of a pattern in a string column

    Args:
        df (pandas.DataFrame): data passed in through the pipe.
        *args: Logical conditions

    Example:
        ## Setup
        import grama as gr
        DF = gr.Intention()
        ## Load example dataset
        from grama.data import df_diamonds

        ## Apply some filters
        (
            df_diamonds
            >> gr.tf_filter(
                ## Remove invalid dimensions
                0 < DF.x,
                0 < DF.y,
                0 < DF.z,
                ## Remove missing values
                gr.not_nan(DF.carat),
            )
        )

    """
    mask = Series(ones(df.shape[0], dtype=bool))
    for arg in args:
        if arg.dtype != bool:
            raise Exception("Arguments must be boolean.")
        mask = mask & arg.reset_index(drop=True)
    return df[mask.values].reset_index(drop=True)

tran_filter = mask  # alias for mask()

tf_filter = add_pipe(tran_filter)


@dfdelegate
def tran_top_n(df, n=None, ascending=True, col=None):
    if not n:
        raise ValueError("n must be specified")
    if not isinstance(col, Series):
        col = df.columns[-1]
    else:
        col = col._name
    index = df[[col]].copy()
    index["ranks"] = index[col].rank(ascending=ascending)
    index = index[index["ranks"] >= index["ranks"].nlargest(n).min()]
    return df.reindex(index.index)

tf_top_n = add_pipe(tran_top_n)


@dfdelegate
def tran_pull(df, col=-1):
    if not isinstance(col, Series):
        col = df.columns[-1]
    else:
        col = col._name

    return df[col]

tf_pull = add_pipe(tran_pull)


@dfdelegate
def tran_dropna(df, how="any", subset=None):
   return df.dropna(how=how, subset=subset).reset_index(drop=True)

tf_dropna = add_pipe(tran_dropna)



import numpy as np
import pandas as pd
import unittest
import io
import sys

from context import grama as gr
from context import data
from context import models

DF = gr.Intention()

## Test transform tools
##################################################
class TestTools(unittest.TestCase):
    def setUp(self):
        pass

    def test_tran_md(self):
        md = models.make_test()

        ## Check for identical responses
        df = gr.df_make(x0=1, x1=1, x2=1)
        df_ev = gr.eval_df(md, df=df)
        df_tf = gr.tran_md(df, md=md)

        self.assertTrue(gr.df_equal(df_ev, df_tf))

    def test_bootstrap(self):
        df_stang = data.df_stang
        df_stang._meta = "foo"

        def tran_stats(df):
            val = df.select_dtypes(include="number").values

            means = np.mean(val, axis=0)
            stds = np.std(val, axis=0)

            # Check metadata propagation
            self.assertTrue(df._meta == "foo")

            return pd.DataFrame(
                data={
                    "var": df.select_dtypes(include="number").columns,
                    "mean": means,
                    "std": stds,
                }
            )

        df_res = gr.tran_bootstrap(
            df_stang, tran=tran_stats, n_boot=3e0, n_sub=3e0, seed=101
        )

        df_sel = gr.tran_bootstrap(
            df_stang, tran=tran_stats, n_boot=3e0, n_sub=3e0, seed=101, col_sel=["mean"]
        )

        df_piped = df_stang >> gr.tf_bootstrap(
            tran=tran_stats, n_boot=3e0, n_sub=3e0, seed=101
        )

        ## Test output shape
        self.assertTrue(
            set(df_res.columns)
            == set(["var", "mean", "mean_lo", "mean_up", "std", "std_lo", "std_up"])
        )
        self.assertTrue(df_res.shape[0] == 4)

        self.assertTrue(
            set(df_sel.columns) == set(["var", "mean", "mean_lo", "mean_up", "std"])
        )
        self.assertTrue(df_sel.shape[0] == 4)

        ## Test pipe
        self.assertTrue(gr.df_equal(df_res, df_piped))

    def test_outer(self):
        df = pd.DataFrame(dict(x=[1, 2]))
        df_outer = pd.DataFrame(dict(y=[3, 4]))

        df_true = pd.DataFrame(dict(x=[1, 2, 1, 2], y=[3, 3, 4, 4]))

        df_res = gr.tran_outer(df, df_outer)
        df_piped = df >> gr.tf_outer(df_outer)

        pd.testing.assert_frame_equal(
            df_true,
            df_res,
            check_exact=False,
            check_dtype=False,
            check_column_type=False,
        )
        pd.testing.assert_frame_equal(
            df_piped,
            df_res,
            check_exact=False,
            check_dtype=False,
            check_column_type=False,
        )

        # Empty cases
        gr.df_equal(
            df,
            gr.tran_outer(pd.DataFrame(), df_outer=df)
        )
        gr.df_equal(
            df,
            gr.tran_outer(df, df_outer=pd.DataFrame())
        )

    def test_gauss_copula(self):
        md = gr.Model() >> gr.cp_marginals(
            E=gr.marg_fit("norm", data.df_stang.E),
            mu=gr.marg_fit("beta", data.df_stang.mu),
            thick=gr.marg_fit("uniform", data.df_stang.thick),
        )
        df_corr = gr.tran_copula_corr(data.df_stang, model=md)

    def test_kfold(self):
        df_train = pd.DataFrame(
            dict(X=list(range(4)), Y=[0, 0, 1, 1], _fold=["a", "a", "b", "b"])
        )
        df_true = pd.DataFrame(
            dict(
                mse_Y=[
                    np.mean((np.array([0, 0]) - 1) ** 2),
                    np.mean((np.array([1, 1]) - 0) ** 2),
                ],
                _kfold=[0, 1],
            )
        )
        df_true_m = pd.DataFrame(
            dict(
                mse_Y=[
                    np.mean((np.array([0, 0]) - 1) ** 2),
                    np.mean((np.array([1, 1]) - 0) ** 2),
                ],
                _fold=["a", "b"],
            )
        )

        ## Unshuffled, auto-generated folds
        df_res = df_train >> gr.tf_kfolds(
            k=2,
            ft=gr.ft_rf(out=["Y"], var=["X"]),
            shuffle=False,
            summaries=dict(mse=gr.mse),
        )

        pd.testing.assert_frame_equal(
            df_res,
            df_true,
            check_exact=False,
            check_dtype=False,
            check_column_type=False,
        )

        ## Manual folds
        df_manual = df_train >> gr.tf_kfolds(
            ft=gr.ft_rf(out=["Y"], var=["X"]),
            var_fold="_fold",
            summaries=dict(mse=gr.mse),
        )

        pd.testing.assert_frame_equal(
            df_manual,
            df_true_m,
            check_exact=False,
            check_dtype=False,
            check_column_type=False,
        )


# --------------------------------------------------
class TestSummaries(unittest.TestCase):
    def setUp(self):
        self.md = models.make_test()

    def test_sobol(self):
        ## Minimal
        df_first = gr.eval_hybrid(self.md, df_det="nom")
        df_sobol = gr.tran_sobol(df_first)
        self.assertTrue(set(df_sobol.columns) == set(["y0", "ind"]))
        self.assertTrue(set(df_sobol["ind"]) == set(["S_x0", "S_x1"]))

        ## Full
        df_full = gr.tran_sobol(df_first, full=True)
        self.assertTrue(set(df_full.columns) == set(["y0", "ind"]))
        self.assertTrue(
            set(df_full["ind"]) == set(["S_x0", "S_x1", "T_x0", "T_x1", "var"])
        )

        ## Total order
        df_total = gr.eval_hybrid(self.md, df_det="nom", plan="total")
        df_sobol_total = gr.tran_sobol(df_total)
        self.assertTrue(set(df_sobol.columns) == set(["y0", "ind"]))
        self.assertTrue(set(df_sobol["ind"]) == set(["S_x0", "S_x1"]))

    def test_pca(self):
        df_test = pd.DataFrame(dict(x0=[1, 2, 3], x1=[1, 2, 3]))
        df_offset = pd.DataFrame(dict(x0=[1, 2, 3], x1=[3, 4, 5]))
        df_scaled = pd.DataFrame(dict(x0=[1, 2, 3], x1=[1, 3, 5]))

        df_true = pd.DataFrame(
            dict(
                lam=[2, 0],
                x0=[1 / np.sqrt(2), 1 / np.sqrt(2)],
                x1=[1 / np.sqrt(2), -1 / np.sqrt(2)],
            )
        )

        ## Check correctness
        df_pca = df_test >> gr.tf_pca()

        self.assertTrue(np.allclose(df_true.lam, df_pca.lam))

        ## Offset data should not affect results
        df_pca_off = df_offset >> gr.tf_pca()
        self.assertTrue(gr.df_equal(df_pca, df_pca_off))

        ## Check standardized form
        df_pca_scaled = df_test >> gr.tf_pca(standardize=True)
        self.assertTrue(gr.df_equal(df_pca, df_pca_scaled))

    def test_iocorr(self):
        df = (
            gr.df_make(x=[1., 2., 3., 4.])
            >> gr.tf_mutate(
                y=+0.5 * DF.x,
                z=-0.5 * DF.x,
            )
            >> gr.tf_iocorr(var=["x"], out=["y", "z"])
        )
        df_true = gr.df_make(
            var=["x", "x"],
            out=["y", "z"],
            rho=[1.0, -1.0],
        )

        ## Check for correct values
        self.assertTrue(gr.df_equal(df, df_true))


# --------------------------------------------------
class TestAsub(unittest.TestCase):
    def setUp(self):
        pass

    def test_asub(self):
        df_data = pd.DataFrame(
            dict(Df_Dx=[1 / np.sqrt(2)] * 2, Df_Dy=[1 / np.sqrt(2)] * 2)
        )
        df_true = pd.DataFrame(
            dict(
                x=[+1 / np.sqrt(2), +1 / np.sqrt(2)],
                y=[+1 / np.sqrt(2), -1 / np.sqrt(2)],
                out=["f", "f"],
                lam=[1, 0],
            )
        )

        df_res = gr.tran_asub(df_data)

        ## Entries correct
        angles = gr.tran_angles(df_true[["x", "y"]], df_res[["x", "y"]])
        self.assertTrue(np.allclose(angles, [0, 0]))

        ## Expected columns
        self.assertTrue(set(df_res.columns) == set(df_true.columns))

        df_piped = df_data >> gr.tf_asub()
        self.assertTrue(df_res.equals(df_piped))


# --------------------------------------------------
class TestAngles(unittest.TestCase):
    def setUp(self):
        self.df = pd.DataFrame(dict(v=[1, 1]))
        self.df_v1 = pd.DataFrame(dict(w=[+1, -1]))
        self.df_v2 = pd.DataFrame(dict(w=[+1, +1]))

    def test_angles(self):
        theta1 = gr.tran_angles(self.df, self.df_v1)
        theta2 = gr.tran_angles(self.df, self.df_v2)

        self.assertTrue(np.isclose(theta1, np.pi / 2))
        self.assertTrue(np.isclose(theta2, 0))

        theta_piped = self.df >> gr.tf_angles(self.df_v1)
        self.assertTrue(np.isclose(theta1, theta_piped))

        with self.assertRaises(ValueError):
            gr.tran_angles(pd.DataFrame(), self.df_v1)

        with self.assertRaises(ValueError):
            gr.tran_angles(self.df, pd.DataFrame())


# --------------------------------------------------
class TestInner(unittest.TestCase):
    def setUp(self):
        pass

    def test_inner(self):
        df = pd.DataFrame(dict(v=[+1, +1], w=[-1, +1]))

        df_w = pd.DataFrame(dict(v=[1, 0, 1], w=[0, 1, 1], id=["v", "w", "x"]))
        df_true = df.copy()
        df_true["dot_v"] = df.v
        df_true["dot_w"] = df.w
        df_true["dot_x"] = df.v + df.w

        ## Core functionality
        df_noappend = gr.tran_inner(df, df_w, name="id", append=False)

        pd.testing.assert_frame_equal(
            df_true[["dot_v", "dot_w", "dot_x"]],
            df_noappend,
            check_exact=False,
            check_dtype=False,
            check_column_type=False,
        )

        df_res = gr.tran_inner(df, df_w, name="id")
        pd.testing.assert_frame_equal(
            df_true,
            df_res,
            check_exact=False,
            check_dtype=False,
            check_column_type=False,
        )

        ## Pipe
        df_piped = df >> gr.tf_inner(df_w, name="id")
        self.assertTrue(df_res.equals(df_piped))

        ## Small weights
        df_w_small = pd.DataFrame(dict(v=[1], w=[1]))
        df_true_small = df.copy()
        df_true_small["dot"] = df.v + df.w

        df_res_small = gr.tran_inner(df, df_w_small)
        pd.testing.assert_frame_equal(
            df_true_small,
            df_res_small,
            check_exact=False,
            check_dtype=False,
            check_column_type=False,
        )

        ## No name column
        df_w_noname = df_w.drop("id", axis=1)
        df_true_noname = df.copy()
        df_true_noname["dot0"] = df.v
        df_true_noname["dot1"] = df.w
        df_true_noname["dot2"] = df.v + df.w

        df_res_noname = gr.tran_inner(df, df_w_noname)
        pd.testing.assert_frame_equal(
            df_true_noname,
            df_res_noname,
            check_exact=False,
            check_dtype=False,
            check_column_type=False,
        )


# --------------------------------------------------
class TestDR(unittest.TestCase):
    def setUp(self):
        pass

    def test_tsne(self):
        ## t-SNE executes successfully
        df_tsne = (
            data.df_diamonds >> gr.tf_sample(n=100) >> gr.tf_tsne(var=["x", "y", "z"])
        )

    def test_umap(self):
        ## UMAP executes successfully
        df_umap = (
            data.df_diamonds >> gr.tf_sample(n=100) >> gr.tf_umap(var=["x", "y", "z"])
        )


# --------------------------------------------------
# class TestFeaturize(unittest.TestCase):
#     def setUp(self):
#         pass

#     def test_tran_poly(self):
#         df = gr.df_make(x=[0.0, 1.0, 0.0], y=[0.0, 0.0, 1.0], z=[1.0, 2.0, 3.0],)
#         df_true = df.copy()
#         df_true["1"] = [1.0, 1.0, 1.0]
#         df_true["x^2"] = [0.0, 1.0, 0.0]
#         df_true["x y"] = [0.0, 0.0, 0.0]
#         df_true["y^2"] = [0.0, 0.0, 1.0]

#         df_res = gr.tran_poly(df, var=["x", "y"], degree=2, keep=True)
#         self.assertTrue(gr.df_equal(df_true, df_res[df_true.columns]))


# --------------------------------------------------
# class TestMatminer(unittest.TestCase):
#     def test_magpie_featurizer(self):
#         df_magpie_check = pd.DataFrame(
#             {
#                 "MagpieData minimum Number": [1.0],
#                 "MagpieData maximum Number": [8.0],
#                 "MagpieData range Number": [7.0],
#             }
#         )
#         df_test = gr.df_make(FORMULA=["C6H12O6"])

#         df_res = df_test >> gr.tf_feat_composition()

#         self.assertTrue(gr.df_equal(df_test, df_res[df_test.columns]))



import numpy as np
import pandas as pd
from scipy.stats import norm
import unittest

from context import grama as gr
from context import data

X = gr.Intention()

##==============================================================================
## reshape test functions
##==============================================================================
def arrange_apply_helperfunc(df):
    df = df.sort_values("depth", ascending=False)
    df = df.head(5)
    return df


class TestReshape(unittest.TestCase):
    def setUp(self):
        self.df_elongated = data.df_diamonds >> gr.tf_gather(
            "variable", "value", add_id=True
        )

    def test_arrange(self):
        df = (
            data.df_diamonds.groupby("cut")
            .apply(arrange_apply_helperfunc)
            .reset_index(drop=True)
        )
        d = (
            data.df_diamonds
            >> gr.tf_group_by("cut")
            >> gr.tf_arrange("depth", ascending=False)
            >> gr.tf_head(5)
            >> gr.tf_ungroup()
        ).reset_index(drop=True)
        self.assertTrue(df.equals(d))

        d = (
            data.df_diamonds
            >> gr.tf_group_by("cut")
            >> gr.tf_arrange(X.depth, ascending=False)
            >> gr.tf_head(5)
            >> gr.tf_ungroup()
        ).reset_index(drop=True)
        assert df.equals(d)

        df = data.df_diamonds.sort_values(["cut", "price"], ascending=False) \
                             .reset_index(drop=True)
        d = data.df_diamonds >> gr.tf_arrange(gr.desc(X.cut), gr.desc(X.price))
        self.assertTrue(df.equals(d))

    def test_rename(self):
        df = data.df_diamonds.rename(
            columns={"cut": "Cut", "table": "Table", "carat": "Carat"}
        )
        d = data.df_diamonds >> gr.tf_rename(Cut=X.cut, Table=X.table, Carat="carat")
        self.assertTrue(df.equals(d))

    def test_gather(self):
        d = data.df_diamonds >> gr.tf_gather(
            "variable", "value", ["price", "depth", "x", "y", "z"]
        )

        variables = ["price", "depth", "x", "y", "z"]
        id_vars = [c for c in data.df_diamonds.columns if c not in variables]
        df = pd.melt(data.df_diamonds, id_vars, variables, "variable", "value")

        self.assertTrue(df.equals(d))

        d = data.df_diamonds >> gr.tf_gather("variable", "value")

        variables = data.df_diamonds.columns.tolist()
        id_vars = []
        df = pd.melt(data.df_diamonds, id_vars, variables, "variable", "value")

        self.assertTrue(df.equals(d))

        df = data.df_diamonds.copy()
        df["_ID"] = np.arange(df.shape[0])
        df = pd.melt(df, ["_ID"], variables, "variable", "value")

        self.assertTrue(df.equals(self.df_elongated))

    def test_spread(self):
        columns = self.df_elongated.columns.tolist()
        id_cols = ["_ID"]

        df = self.df_elongated.copy()
        df["temp_index"] = df["_ID"].values
        df = df.set_index("temp_index")
        spread_data = df[["variable", "value"]]

        spread_data = spread_data.pivot(columns="variable", values="value")
        converted_spread = spread_data.copy()

        columns_to_convert = [col for col in spread_data if col not in columns]
        converted_spread = gr.convert_type(converted_spread, columns_to_convert)

        df = df[["_ID"]].drop_duplicates()

        df_spread = df.merge(
            spread_data, left_index=True, right_index=True
        ).reset_index(drop=True)
        df_conv = df.merge(
            converted_spread, left_index=True, right_index=True
        ).reset_index(drop=True)

        d_spread = self.df_elongated >> gr.tf_spread("variable", "value")
        d_spread_conv = self.df_elongated >> gr.tf_spread(
            "variable", "value", convert=True
        )

        self.assertTrue(df_spread.equals(d_spread))
        self.assertTrue(df_conv.equals(d_spread_conv))

        ## Test fill
        df_base = gr.df_make(
            x=[1, 2, 3, 4, 5],
            y=["a", "b", "c", "a", "b"],
            idx=[0, 0, 0, 1, 1]
        )
        df_true = gr.df_make(
            a=[1, 4],
            b=[2, 5],
            c=[3, 0],
            idx=[0, 1]
        )
        df_res = df_base >> gr.tf_spread(X.y, X.x, fill=0)

        self.assertTrue(gr.df_equal(df_true, df_res, close=True))

    def test_separate(self):

        d = pd.DataFrame({"a": ["1-a-3", "1-b", "1-c-3-4", "9-d-1", "10"]})

        test1 = d >> gr.tf_separate(
            X.a,
            ["a1", "a2", "a3"],
            remove=True,
            convert=False,
            extra="merge",
            fill="right",
        )

        true1 = pd.DataFrame(
            {
                "a1": ["1", "1", "1", "9", "10"],
                "a2": ["a", "b", "c", "d", np.nan],
                "a3": ["3", np.nan, "3-4", "1", np.nan],
            }
        )
        self.assertTrue(true1.equals(test1))

        test2 = d >> gr.tf_separate(
            X.a,
            ["a1", "a2", "a3"],
            remove=True,
            convert=False,
            extra="merge",
            fill="left",
        )

        true2 = pd.DataFrame(
            {
                "a1": ["1", np.nan, "1", "9", np.nan],
                "a2": ["a", "1", "c", "d", np.nan],
                "a3": ["3", "b", "3-4", "1", "10"],
            }
        )
        self.assertTrue(true2.equals(test2))

        test3 = d >> gr.tf_separate(
            X.a,
            ["a1", "a2", "a3"],
            remove=True,
            convert=True,
            extra="merge",
            fill="right",
        )

        true3 = pd.DataFrame(
            {
                "a1": [1, 1, 1, 9, 10],
                "a2": ["a", "b", "c", "d", np.nan],
                "a3": ["3", np.nan, "3-4", "1", np.nan],
            }
        )
        self.assertTrue(true3.equals(test3))

        test4 = d >> gr.tf_separate(
            X.a,
            ["col1", "col2"],
            sep=[1, 3],
            remove=True,
            convert=False,
            extra="drop",
            fill="left",
        )

        true4 = pd.DataFrame(
            {"col1": ["1", "1", "1", "9", "1"], "col2": ["-a", "-b", "-c", "-d", "0"]}
        )
        self.assertTrue(true4.equals(test4))

        test5 = d >> gr.tf_separate(
            X.a,
            ["col1", "col2"],
            sep=[1, 3],
            remove=False,
            convert=False,
            extra="drop",
            fill="left",
        )

        true5 = pd.DataFrame(
            {
                "a": ["1-a-3", "1-b", "1-c-3-4", "9-d-1", "10"],
                "col1": ["1", "1", "1", "9", "1"],
                "col2": ["-a", "-b", "-c", "-d", "0"],
            }
        )
        self.assertTrue(true5.equals(test5))

        test6 = d >> gr.tf_separate(
            X.a,
            ["col1", "col2", "col3"],
            sep=[30],
            remove=True,
            convert=False,
            extra="drop",
            fill="left",
        )

        true6 = pd.DataFrame(
            {
                "col1": ["1-a-3", "1-b", "1-c-3-4", "9-d-1", "10"],
                "col2": [np.nan, np.nan, np.nan, np.nan, np.nan],
                "col3": [np.nan, np.nan, np.nan, np.nan, np.nan],
            }
        )
        self.assertTrue(true6.equals(test6))

    def test_unite(self):
        d = pd.DataFrame(
            {"a": [1, 2, 3], "b": ["a", "b", "c"], "c": [True, False, np.nan]}
        )

        test1 = d >> gr.tf_unite(
            "united", X.a, "b", 2, remove=True, na_action="maintain"
        )
        true1 = pd.DataFrame({"united": ["1_a_True", "2_b_False", np.nan]})
        self.assertTrue(true1.equals(test1))

        test2 = d >> gr.tf_unite(
            "united", ["a", "b", "c"], remove=True, na_action="ignore", sep="*"
        )
        true2 = pd.DataFrame({"united": ["1*a*True", "2*b*False", "3*c"]})
        self.assertTrue(test2.equals(true2))

        test3 = d >> gr.tf_unite(
            "united", d.columns, remove=True, na_action="as_string"
        )
        true3 = pd.DataFrame({"united": ["1_a_True", "2_b_False", "3_c_nan"]})
        self.assertTrue(true3.equals(test3))

        test4 = d >> gr.tf_unite(
            "united", d.columns, remove=False, na_action="as_string"
        )
        true4 = pd.DataFrame(
            {
                "a": [1, 2, 3],
                "b": ["a", "b", "c"],
                "c": [True, False, np.nan],
                "united": ["1_a_True", "2_b_False", "3_c_nan"],
            }
        )

        self.assertTrue(true4.equals(test4))

class TestNesting(unittest.TestCase):
    def setUp(self):
        pass

    def test_explode(self):
        df_base = gr.df_make(x=[1, 2], y=[[3, 4], [5, 6]])
        df_str = gr.df_make(x=[1, 2], y=[["3", "4"], ["5", "6"]])
        df_true = gr.df_make(
            x=[1, 1, 2, 2],
            y=[3, 4, 5, 6]
        )

        df_res = df_base >> gr.tf_explode(X.y)
        df_res_s = df_base >> gr.tf_explode(X.y, convert=True)

        self.assertTrue(gr.df_equal(df_true, df_res, close=True))
        self.assertTrue(gr.df_equal(df_true, df_res_s, close=True))
